from copy import deepcopy
from ahc.Ahc import ComponentModel
from ahc.Ahc import ConnectorTypes
from ahc.Ahc import Event
from ahc.Ahc import EventTypes
from ahc.Ahc import GenericMessage
from ahc.Ahc import GenericMessageHeader
from ahc.Routing.DSR.ApplicationComponent import ApplicationComponent
from ahc.Routing.DSR.DSRAlgorithmComponent import DSRAlgorithmComponent
from ahc.Routing.DSR.DSRAlgorithmComponent import MessageTypes


class AdhocNodeComponent(ComponentModel):
    def __init__(self, component_name, component_id):
        super(AdhocNodeComponent, self).__init__(component_name, component_id)
        self.Application = ApplicationComponent(ApplicationComponent.__name__, component_id)
        self.DSRAlgorithmComponent = DSRAlgorithmComponent(DSRAlgorithmComponent.__name__, component_id)

        self.Application.connect_me_to_component(ConnectorTypes.DOWN, self.DSRAlgorithmComponent)
        self.DSRAlgorithmComponent.connect_me_to_component(ConnectorTypes.UP, self.Application)

        self.DSRAlgorithmComponent.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.DSRAlgorithmComponent)

    def send_unicast_message(self, next_hop, payload):
        src = str(self.componentinstancenumber)
        dst = str(next_hop)
        header = GenericMessageHeader("MessageTypeHere",
                                      self.componentname + "-" + src,
                                      self.componentname + "-" + dst,
                                      interfaceid=src + "-" + dst)

        message = GenericMessage(header, payload)

        event = Event(self, EventTypes.MFRT, message)
        self.send_down(event)

    def send_broadcast_message(self, payload):
        for interface in self.connectors[ConnectorTypes.DOWN]:
            for next_hop in interface.componentinstancenumber.split("-"):
                if int(next_hop) != self.get_component_id():
                    self.send_unicast_message(next_hop, payload)

    def send_message(self, next_hop, payload):
        if str("inf") == next_hop:
            self.send_broadcast_message(payload)
        else:
            self.send_unicast_message(next_hop, payload)

    def get_next_component_id(self, route: list):
        try:
            index_of_current_component = route.index(self.get_component_id())
            index_of_next_component = index_of_current_component + 1
            return route[index_of_next_component]
        except ValueError:
            print("[AdhocNodeComponent:get_next_component_id][Exception] ValueError")
            print("[AdhocNodeComponent:get_next_component_id][Exception] comp_id = " + str(self.get_component_id()))
            str_route = ' '.join(map(str, route))
            print("[AdhocNodeComponent:get_next_component_id][Exception] route = " + str_route)
            return None

    def get_prev_component_id(self, route: list):
        try:
            index_of_current_component = route.index(self.get_component_id())
            index_of_prev_component = index_of_current_component - 1
            return route[index_of_prev_component]
        except ValueError:
            print("[AdhocNodeComponent:get_prev_component_id][Exception] ValueError")
            print("[AdhocNodeComponent:get_prev_component_id][Exception] comp_id = " + str(self.get_component_id()))
            str_route = ' '.join(map(str, route))
            print("[AdhocNodeComponent:get_prev_component_id][Exception] route = " + str_route)
            return None

    def get_component_id(self) -> int:
        return self.componentinstancenumber

    def on_message_from_bottom(self, eventobj: Event):

        message_type = eventobj.eventcontent.payload.header.messagetype
        message_from = eventobj.eventcontent.payload.header.messagefrom
        message_to = eventobj.eventcontent.payload.header.messageto

        message_header = GenericMessageHeader(message_type,
                                              message_from,
                                              message_to)

        message_payload = eventobj.eventcontent.payload.payload
        message = GenericMessage(message_header, message_payload)

        event = Event(self, EventTypes.MFRB, message)
        self.send_up(event)

    def is_link_up(self, next_hop) -> bool:

        for interface in self.connectors[ConnectorTypes.DOWN]:
            if interface.componentinstancenumber.split("-")[0] == str(self.componentinstancenumber):
                interface_next_hop = interface.componentinstancenumber.split("-")[1]
                if interface_next_hop == str(next_hop):
                    return True
            else:
                interface_next_hop = interface.componentinstancenumber.split("-")[0]
                if interface_next_hop == str(next_hop):
                    return True

        return False

    def on_message_from_top(self, eventobj: Event):

        src = int(eventobj.eventcontent.header.messagefrom.split("-")[1])

        message_type = eventobj.eventcontent.header.messagetype
        route = deepcopy(eventobj.eventcontent.payload[0])

        next_hop = None
        if MessageTypes.ROUTE_REQUEST == message_type:
            next_hop = str("inf")
        elif MessageTypes.ROUTE_REPLY == message_type:
            next_hop = self.get_prev_component_id(route)
        elif MessageTypes.ROUTE_ERROR == message_type:
            next_hop = self.get_prev_component_id(route)
        elif MessageTypes.ROUTE_FORWARDING == message_type:
            next_hop = self.get_next_component_id(route)

        if next_hop is not None:
            self.send_message(next_hop, eventobj.eventcontent)

        if MessageTypes.ROUTE_FORWARDING == message_type:
            if not self.is_link_up(next_hop):
                self.start_route_error_message(src, next_hop, route)

    def start_route_error_message(self, src, next_hop, route):

        try:
            new_src = self.get_component_id()
            new_dst = src

            header = GenericMessageHeader(MessageTypes.ROUTE_ERROR,
                                          self.componentname + "-" + str(new_src),
                                          self.componentname + "-" + str(new_dst))

            index_of_current_component = route.index(self.get_component_id())
            new_route = route[0: index_of_current_component]

            broken_link = [self.get_component_id(), next_hop]

            event = Event(self,
                          EventTypes.MFRB,
                          GenericMessage(header,
                                         [new_route, broken_link]))

            self.send_up(event)

        except ValueError:
            print("[AdhocNodeComponent:send_route_error][Error] component_id not in the route")
            print("[AdhocNodeComponent:send_route_error][Error] comp_id = " + str(self.get_component_id()))
            str_route = ' '.join(map(str, route))
            print("[AdhocNodeComponent:send_route_error][Error] route = " + str_route)



from ahc.Ahc import ComponentModel, Event, ConnectorTypes, EventTypes, Topology
from ahc.Routing.SSBR.ApplicationAndNetworkComponent import ApplicationAndNetwork
from ahc.Routing.SSBR.FPComponent import FP
from ahc.Routing.SSBR.NetworkInterfaceComponent import NetworkInterface
from ahc.Routing.SSBR.DRPComponent import DRP
from ahc.Routing.SSBR.HelperFunctions import messageParser, sendMessageToOtherNode

# Encapsulator for the SSBR Node
class SSBRNode(ComponentModel):
    def __init__(self, componentname, componentid):
        super(SSBRNode, self).__init__(componentname, componentid)
        self.componentid = componentid
        self.ApplicationAndNetwork = ApplicationAndNetwork(ApplicationAndNetwork.__name__, componentid)
        self.FP = FP(FP.__name__, componentid)
        self.DRP = DRP(DRP.__name__, componentid)
        self.NetworkInterface = NetworkInterface(NetworkInterface.__name__, componentid)
        self.messageFrom = -1
        
        #   Building SSBR Connections
        self.ApplicationAndNetwork.connect_me_to_component(ConnectorTypes.DOWN, self.FP)
        
        self.FP.connect_me_to_component(ConnectorTypes.UP, self.ApplicationAndNetwork)
        self.FP.connect_me_to_component(ConnectorTypes.DOWN, self.NetworkInterface)

        self.DRP.connect_me_to_component(ConnectorTypes.PEER, self.FP)

        self.NetworkInterface.connect_me_to_component(ConnectorTypes.UP, self.DRP)

        self.NetworkInterface.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.NetworkInterface)

    def on_init(self, eventobj: Event):
        self.neighbors = Topology().get_neighbors(self.componentinstancenumber) #get neighbours
        print(f"{self.componentname} - #{self.componentid} is up.\n")
        pass

    def on_message_from_bottom(self, eventobj: Event):
       
        evt = Event(self, EventTypes.MFRB,messageParser(self,eventobj))    
        self.send_up(evt) # send incoming messages to upper components

    def on_message_from_top(self, eventobj: Event):
        if eventobj.eventcontent.header.interfaceid != float("inf") and self.messageFrom == -1:
                messageFrom = int(eventobj.eventcontent.header.interfaceid.split("-")[0])
                if int(eventobj.eventcontent.header.interfaceid.split("-")[0]) == int(self.componentinstancenumber):
                    messageFrom = int(eventobj.eventcontent.header.interfaceid.split("-")[1])
                self.messageFrom = messageFrom
        if eventobj.eventcontent.header.messagetype == "ROUTESEARCH":
            
            for neigh in self.neighbors:
                if(int(self.messageFrom) != int(neigh)):
                    evt = Event(self, EventTypes.MFRT,sendMessageToOtherNode(self, eventobj, neigh))
                    self.send_down(evt)   # send incoming messages from upper components to a channel
        elif eventobj.eventcontent.header.messagetype == "ROUTEREPLY":
            messageTo = self.messageFrom
            evt = Event(self, EventTypes.MFRT,sendMessageToOtherNode(self, eventobj, messageTo))
            self.send_down(evt)
        elif eventobj.eventcontent.header.messagetype == "UNICASTDATA":
            messageTo = eventobj.eventcontent.header.nexthop
            evt = Event(self, EventTypes.MFRT,sendMessageToOtherNode(self, eventobj, messageTo))
            self.send_down(evt)
       




import time
from datetime import datetime
from enum import Enum

from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage
import copy

class DBR2PMesageType(Enum):
    RDRequest = "RouteDiscoveryRequest"
    RDReply = "RouteDiscoveryReply"
    BSP = "BackupSetupPacket"
    LFM = "LinkFailMessage"
    LRM = "LinkRecoveryMessage"
    ERROR = 'error'
    DT = "DataTransfer"  # this is just for testing


# define your own message header structure
class DBR2PMessageHeader(GenericMessageHeader):
    def __init__(self, messagetype, messagefrom, messageto, nexthop=float('inf'), interfaceid=float('inf'),
                 sequencenumber=-1):
        super(DBR2PMessageHeader, self).__init__(messagetype, messagefrom, messageto, nexthop, interfaceid,
                                                 sequencenumber)


class DBR2PMessagePayload(GenericMessagePayload):
    def __init__(self, messagepayload, time_now=time.time()):
        super(DBR2PMessagePayload, self).__init__(messagepayload)
        self.time = time_now  # Take time on arrival at destination






import queue
import threading
import time

from ahc.Ahc import ComponentModel, Event,  EventTypes, GenericMessage
from ahc.Routing.DBR2P.Messages import DBR2PMessageHeader, DBR2PMessagePayload, DBR2PMesageType

TIME_TO_WAIT = 2.0  # This is time out in sec for  SOURCE_NODE_ROUTE_DISCOVERY and DESTINATION_RECEIVE_request
Tc_TO_WAIT = 2.0  # This is time to wait in sec to get another RD_request from the same source for intermediate_node_receive_RDrequest
N = 3  # This is for RECEIVE_RD-request to check number of time #RD received


# TOSEC = 1000.0
# TOMS = 1000.0

class RoutingDBR2PComponent(ComponentModel):

    def __init__(self, componentname, componentid):
        super(RoutingDBR2PComponent, self).__init__(componentname, componentid)
        self.SequenceNumber = 1
        self.RDRequestCache = {}  # Key Sequence  Value n For intermediate_node_receive_RDrequest
        self.BackupRouteCache = {}  # Key (Source-Destination) Value Route
        self.TimersIntermediate = {}  # Key Source ID Value Time
        self.TimersDestination = {}  # Key Source ID Value Time
        self.RouteDiscoveryThreads = {}  # key Destination value Condition To Use later on Event to notify Thread
        self.DestinationReceiverThreads = {}  # key Source value Condition To Use later on Event to notify Thread
        self.RDReplies = {}  # key Destination value message for source_node_route_discovery
        self.RDRequests = {}  # key Source value message Queue This is only to put RDRequest if this node is the Destination for threading purposes
        self.LinkMaintenanceThreads = {}  # key (source-destination) Value condition
        self.DTPackets = {}  # key (Source-Destination) Value Queue

    def on_message_from_bottom(self, eventobj: Event):
        # print(f"{EventTypes.MFRB} {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.interfaceid} & {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")
        if eventobj.eventcontent.header.messagetype == DBR2PMesageType.DT:
            if self.componentinstancenumber == eventobj.eventcontent.header.messageto:
                self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))
                self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
            elif self.componentinstancenumber in eventobj.eventcontent.payload.messagepayload:
                key = (eventobj.eventcontent.header.messagefrom, eventobj.eventcontent.header.messageto)
                try:
                    condition = self.LinkMaintenanceThreads[key]
                    condition.acquire()
                    self.DTPackets[key].put(eventobj)
                    condition.notify()
                    condition.release()
                except KeyError:
                    condition = threading.Condition()
                    self.LinkMaintenanceThreads[key] = condition
                    self.DTPackets[key] = queue.Queue()
                    args = (eventobj, condition)
                    thread = threading.Thread(target=self.link_maintenance, args=args)
                    thread.start()
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.RDRequest:
            if self.componentinstancenumber == eventobj.eventcontent.header.messageto:
                try:
                    condition = self.DestinationReceiverThreads[eventobj.eventcontent.header.messagefrom]
                    condition.acquire()
                    self.RDRequests[eventobj.eventcontent.header.messagefrom].put(eventobj)
                    condition.notify()
                    condition.release()
                except KeyError:
                    if eventobj.eventcontent.header.messagefrom in self.TimersDestination.keys():
                        if time.time() - self.TimersDestination[eventobj.eventcontent.header.messagefrom] < Tc_TO_WAIT:
                            return  # Discard
                        else:
                            self.TimersDestination.pop(eventobj.eventcontent.header.messagefrom)
                    condition = threading.Condition()
                    self.DestinationReceiverThreads[eventobj.eventcontent.header.messagefrom] = condition
                    self.RDRequests[eventobj.eventcontent.header.messagefrom] = queue.Queue()
                    args = (eventobj, eventobj.eventcontent.header.sequencenumber, condition)
                    thread = threading.Thread(target=self.destination_receive_RDrequest, args=args)
                    thread.start()
            else:
                self.intermediate_node_receive_RDrequest(eventobj)

        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.LFM:
            if self.componentinstancenumber == eventobj.eventcontent.header.messageto:
                self.link_fail(eventobj)
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.LRM:
            if self.componentinstancenumber == eventobj.eventcontent.header.messageto:
                self.send_up(eventobj)
            elif self.componentinstancenumber in eventobj.eventcontent.payload.messagepayload:
                if self.componentinstancenumber == eventobj.eventcontent.payload.messagepayload[-1]:
                    return  # discard
                #  It came in upstream
                if eventobj.eventsource.componentinstancenumber == eventobj.eventcontent.payload.messagepayload[
                    eventobj.eventcontent.payload.messagepayload.index(self.componentinstancenumber) + 1]:
                    self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.RDReply:
            if self.componentinstancenumber == eventobj.eventcontent.header.messageto:
                try:
                    condition = self.RouteDiscoveryThreads[eventobj.eventcontent.header.messagefrom]
                    condition.acquire()
                    self.RDReplies[eventobj.eventcontent.header.messagefrom] = eventobj
                    condition.notify()
                    condition.release()
                except KeyError:
                    # print(f"Time ot route discovery {self.componentinstancenumber}")
                    return  # Discard
            elif self.componentinstancenumber in eventobj.eventcontent.payload.messagepayload:
                if self.componentinstancenumber == eventobj.eventcontent.payload.messagepayload[-1]:
                    return  # discard
                #  It came in upstream
                if eventobj.eventsource.componentinstancenumber == eventobj.eventcontent.payload.messagepayload[
                    eventobj.eventcontent.payload.messagepayload.index(self.componentinstancenumber) + 1]:
                    self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.BSP:
            if self.componentinstancenumber == eventobj.eventcontent.header.messageto:
                self.receive_BSpacket(eventobj)
            elif self.componentinstancenumber in eventobj.eventcontent.payload.messagepayload:
                if self.componentinstancenumber == eventobj.eventcontent.payload.messagepayload[-1]:
                    return  # discard
                #  It came in upstream
                if eventobj.eventsource.componentinstancenumber == eventobj.eventcontent.payload.messagepayload[
                    eventobj.eventcontent.payload.messagepayload.index(self.componentinstancenumber) + 1]:
                    # print(f"M {self.componentinstancenumber}: {eventobj.eventsource.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} S {eventobj.eventcontent.header.messagefrom} T {eventobj.eventcontent.header.messageto} R : {eventobj.eventcontent.payload.messagepayload}")
                    self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
        else:
            pass  # discard event

    def on_message_from_top(self, eventobj: Event):
        # print(f"{EventTypes.MFRT}  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")
        if eventobj.eventcontent.header.messagetype == DBR2PMesageType.RDRequest:
            if eventobj.eventcontent.header.messageto in self.RouteDiscoveryThreads.keys():
                return  # discard event it should not happen though
            condition = threading.Condition()
            eventobj.eventcontent.header.sequencenumber = self.SequenceNumber
            self.SequenceNumber += 1
            self.RouteDiscoveryThreads[eventobj.eventcontent.header.messageto] = condition
            args = (eventobj, condition)
            source_node_route_discovery_thread = threading.Thread(target=self.source_node_route_discovery, args=args)
            source_node_route_discovery_thread.start()
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.DT:
            key = (eventobj.eventcontent.header.messagefrom, eventobj.eventcontent.header.messageto)
            try:
                condition = self.LinkMaintenanceThreads[key]
                condition.acquire()
                self.DTPackets[key].put(eventobj)
                condition.notify()
                condition.release()
            except KeyError:
                condition = threading.Condition()
                self.LinkMaintenanceThreads[key] = condition
                self.DTPackets[key] = queue.Queue()
                args = (eventobj, condition)
                thread = threading.Thread(target=self.link_maintenance, args=args)
                thread.start()
        else:
            self.send_down(eventobj)

    # This is thread routine since it is blocking op
    def source_node_route_discovery(self, eventobj: Event, condition: threading.Condition):
        # print(f"source_node_route_discovery:  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")

        # start timer
        # create RD request
        # broadcast RD request
        # Wait for id Reply or Time out
        # if RD reply received return the route
        # else return ERROR

        condition.acquire()
        eventobj.eventcontent.payload.messagepayload.append(self.componentinstancenumber)
        eventobj.eventcontent.payload.time = time.time()
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
        value = condition.wait(TIME_TO_WAIT)  # Wait for RD reply ot time out like while
        if value:
            # Send RD Reply to App

            event = self.RDReplies.pop(eventobj.eventcontent.header.messageto)

            event.event = EventTypes.MFRB
            self.send_up(Event(self, EventTypes.MFRB, event.eventcontent))

        else:

            # return ERROR to App
            error_header = DBR2PMessageHeader(DBR2PMesageType.ERROR, eventobj.eventcontent.header.messagefrom,
                                              eventobj.eventcontent.header.messageto,
                                              sequencenumber=eventobj.eventcontent.header.sequencenumber)
            payload = eventobj.eventcontent.payload
            payload.messagepayload = []
            self.send_up(Event(self, EventTypes.MFRB, GenericMessage(error_header, payload)))

        # remove condition from Dictionary
        self.RouteDiscoveryThreads.pop(eventobj.eventcontent.header.messageto)
        condition.release()
    def intermediate_node_receive_RDrequest(self, eventobj: Event):
        # print(f"intermediate_node_receive_RDrequest:  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")

        sequence_number = eventobj.eventcontent.header.sequencenumber
        source = eventobj.eventcontent.header.messagefrom
        if self.is_RDRequest_recevied_first_time(sequence_number) and not self.check_if_in_RDrequest(eventobj):
            self.RDRequestCache[sequence_number] = 1
            self.TimersIntermediate[source] = time.time()  # Start timer
            eventobj.eventcontent.payload.messagepayload.append(self.componentinstancenumber)
            self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
        else:
            if self.check_if_in_RDrequest(eventobj):
                return
            elif self.RDRequestCache[sequence_number] >= N or (
                    time.time() - self.TimersIntermediate[source] <= Tc_TO_WAIT) if (
                    # Discard request if it is from the same source within the time interval
                    source in self.TimersIntermediate.keys()) else False:
                return
            else:
                # This is not in Paper but it must be
                self.TimersIntermediate[source] = time.time()
                # end
                self.RDRequestCache[sequence_number] += 1
                eventobj.eventcontent.payload.messagepayload.append(self.componentinstancenumber)
                self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
                return

    # This is thread routine since it is blocking op
    def destination_receive_RDrequest(self, eventobj: Event, sequence_number, condition: threading.Condition):
        # print(f"destination_receive_RDrequest:  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")

        # set timer
        # send first route to source
        # while not time out
        # receive more RD reqeust from source
        # if time out backup_node_setup
        eventobj.eventcontent.payload.messagepayload.append(self.componentinstancenumber)
        # print(f"route : {eventobj.eventcontent.payload.messagepayload}")
        Routes = [eventobj.eventcontent.payload.messagepayload]
        start_time = time.time()
        header = DBR2PMessageHeader(DBR2PMesageType.RDReply, self.componentinstancenumber,
                                    eventobj.eventcontent.header.messagefrom, sequencenumber=sequence_number)
        payload = eventobj.eventcontent.payload
        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))
        # A little change at algo
        # take end send reply for the first RDrequest
        # Then start to wait for other

        value = True
        while value:
            condition.acquire()
            time_to_wait = TIME_TO_WAIT - (time.time() - start_time)
            # print( f"Time to wait receiver : {time_to_wait}")
            if time_to_wait > 0:
                value = condition.wait(time_to_wait)  # Wait for TIME sec
            else:
                value = False
            if value:

                event_queue = self.RDRequests[eventobj.eventcontent.header.messagefrom]
                while not event_queue.empty():
                    event = event_queue.get()
                    event.eventcontent.payload.messagepayload.append(self.componentinstancenumber)
                    # print(f"route backup: {event.eventcontent.payload.messagepayload}")
                    Routes.append(event.eventcontent.payload.messagepayload)

            condition.release()

        time_to_wait = TIME_TO_WAIT - (time.time() - start_time)
        # print(f"Time to wait receiver : {time_to_wait}")
        # Remove condition var and Queue
        condition.acquire()
        self.TimersDestination[eventobj.eventcontent.header.messagefrom] = time.time()
        self.RDRequests.pop(eventobj.eventcontent.header.messagefrom)
        self.DestinationReceiverThreads.pop(eventobj.eventcontent.header.messagefrom)
        condition.release()
        # Start BackUp_Node_Setup
        self.backup_node_setup(sequence_number, Routes)
        pass

    def backup_node_setup(self, sequence_number, Routes: list):
        Backup = self.find_backup_node(Routes)
        self.setup_backup(Backup, sequence_number)

    def find_backup_node(self, Routes: list):
        Backup = {}  # key node id Value List of Routes
        for routetmp1 in Routes[:-1]:
            for routetmp2 in Routes[1:]:
                route1 = routetmp1
                route2 = routetmp2
                flag1 = 0
                flag2_start = 0
                length_of_route1 = len(route1)
                length_of_route2 = len(route2)
                while flag1 < length_of_route1:
                    flag2 = flag2_start
                    while flag2 < length_of_route2:
                        if route1[flag1] == route2[flag2]:
                            while flag1 < length_of_route1 or flag2 < length_of_route2:
                                if route1[flag1] == route2[flag2]:
                                    flag1 += 1
                                    flag2 += 1
                                else:
                                    break
                            if (0 < flag1 < length_of_route1) and (0 < flag2 < length_of_route2):
                                bakcup_node_temp = route1[flag1 - 1]
                                bakcup_route_temp = route1[flag1 - 1:]
                                if bakcup_route_temp not in Backup.values():
                                    if bakcup_node_temp not in Backup:
                                        Backup[bakcup_node_temp] = []
                                    if bakcup_route_temp not in Backup[bakcup_node_temp]:
                                        bakcup_route_temp.insert(0,routetmp1[0])
                                        Backup[bakcup_node_temp].append(bakcup_route_temp)
                                bakcup_node_temp = route2[flag2 - 1]
                                bakcup_route_temp = route2[flag2 - 1:]
                                if bakcup_route_temp not in Backup.values():
                                    if bakcup_node_temp not in Backup:
                                        Backup[bakcup_node_temp] = []
                                    if bakcup_route_temp not in Backup[bakcup_node_temp]:
                                        bakcup_route_temp.insert(0, routetmp1[0])
                                        Backup[bakcup_node_temp].append(bakcup_route_temp)
                            flag2_start = flag2
                            flag2 -= 1
                        flag2 += 1
                    flag1 += 1
        return Backup

    def setup_backup(self, BackUp: dict, sequence_number):
        for backupnode in BackUp.keys():
            for backuproute in BackUp[backupnode][1:]:
                header = DBR2PMessageHeader(DBR2PMesageType.BSP, self.componentinstancenumber, backupnode,
                                            sequencenumber=sequence_number)
                payload = DBR2PMessagePayload(backuproute)

                evt = Event(self, EventTypes.MFRT, GenericMessage(header, payload))
                # print(f"send backup:  {self.componentname}.{self.componentinstancenumber} => {evt.eventcontent.header.messagetype} {evt.eventcontent.header.messagefrom} $ {evt.eventcontent.header.messageto} R {evt.eventcontent.payload.messagepayload}")

                self.send_down(evt)

    def receive_BSpacket(self, eventobj: Event):
        # print(
        #     f"BSPacket:  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto} BR :{eventobj.eventcontent.payload.messagepayload}")

        key = (eventobj.eventcontent.payload.messagepayload[0], eventobj.eventcontent.header.messagefrom)
        backup_route = eventobj.eventcontent.payload.messagepayload[1:]
        try:
            if eventobj.eventcontent.payload.messagepayload not in self.BackupRouteCache[key]:
                self.BackupRouteCache[key].append(backup_route)
        except KeyError:
            self.BackupRouteCache[key] = [backup_route]

    # It is a thread routine since it blocking op
    def link_maintenance(self, eventobj: Event, condition: threading.Condition):
        evt = Event(self, EventTypes.MFRT, eventobj.eventcontent)
        times = []
        times.append(time.time())
        self.send_down(evt)
        value = True
        key = (eventobj.eventcontent.header.messagefrom, eventobj.eventcontent.header.messageto)

        while value:
            condition.acquire()
            try:
                time_to_wait = TIME_TO_WAIT - (time.time() - times[0])
            except:
                time_to_wait = TIME_TO_WAIT
            if time_to_wait > 0:
                value = condition.wait(time_to_wait)  # Wait for
            else:
                value = False

            if not value and len(times) != 0:
                eventobj.eventcontent.header.messagetype = DBR2PMesageType.LFM
                eventobj.eventcontent.payload.time = time.time()
                self.link_fail(eventobj)
            else:
                # check if it is a new item then set timer for next one and pre
                # if it is for this one close the thread
                evt: Event

                while not self.DTPackets[key].empty():
                    evt = self.DTPackets[key].get()
                    if evt.eventcontent.payload.messagepayload[evt.eventcontent.payload.messagepayload.index(
                            self.componentinstancenumber) + 1] == evt.eventsource.componentinstancenumber:
                        try:
                            times.pop(0)
                        except:
                            pass
                    else:
                        if self.componentinstancenumber == evt.eventcontent.header.messagefrom:
                            new_evt = Event(self, EventTypes.MFRT, evt.eventcontent)
                            times.append(time.time())
                            self.send_down(new_evt)
                        elif evt.eventcontent.payload.messagepayload[evt.eventcontent.payload.messagepayload.index(
                                self.componentinstancenumber) - 1] == evt.eventsource.componentinstancenumber:
                            new_evt = Event(self, EventTypes.MFRT, evt.eventcontent)
                            times.append(time.time())
                            self.send_down(new_evt)
                condition.release()

        condition.acquire()
        self.LinkMaintenanceThreads.pop(key)
        self.DTPackets.pop(key)
        condition.release()

    def link_fail(self, eventobj: Event):
        # print(
        #     f"link_fail:  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")

        key = (eventobj.eventcontent.payload.messagepayload[0], eventobj.eventcontent.payload.messagepayload[-1])

        if key in self.BackupRouteCache.keys():
            backup_route = self.BackupRouteCache[key].pop(0)
            if not self.BackupRouteCache[key]:  # if no Backup Route remove
                self.BackupRouteCache.pop(key)
            header = DBR2PMessageHeader(DBR2PMesageType.LRM, self.componentinstancenumber,
                                        eventobj.eventcontent.payload.messagepayload[0],
                                        sequencenumber=eventobj.eventcontent.header.sequencenumber)
            messagepayload = DBR2PMessagePayload(eventobj.eventcontent.payload.messagepayload[
                                                 :eventobj.eventcontent.payload.messagepayload.index(
                                                     backup_route[0])] + backup_route)
            messagepayload.time = time.time()
            # if this is source turn data and route to app
            # print(
            #     f"BackROute:  {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto} R: {messagepayload}")

            if eventobj.eventcontent.payload.messagepayload[0] == self.componentinstancenumber:
                self.send_up(Event(self, EventTypes.MFRB, GenericMessage(header, messagepayload)))
            else:
                # if not send to source
                self.send_down(Event(self, EventTypes.MFRT, GenericMessage(header, messagepayload)))

        elif eventobj.eventcontent.payload.messagepayload[0] == self.componentinstancenumber:
            if eventobj.eventcontent.header.messageto in self.RouteDiscoveryThreads.keys():
                return  # discard event
            condition = threading.Condition()
            dest = eventobj.eventcontent.payload.messagepayload[-1]
            if dest not in self.RouteDiscoveryThreads.keys():
                header = DBR2PMessageHeader(DBR2PMesageType.RDRequest, self.componentinstancenumber, dest, sequencenumber=self.SequenceNumber)
                payload = DBR2PMessagePayload([])
                self.SequenceNumber += 1
                self.RouteDiscoveryThreads[dest] = condition
                args = (Event(self, EventTypes.MFRT, GenericMessage(header, payload)), condition)
                source_node_route_discovery_thread = threading.Thread(target=self.source_node_route_discovery, args=args)
                source_node_route_discovery_thread.start()
        else:
            header = DBR2PMessageHeader(DBR2PMesageType.LFM, self.componentinstancenumber,
                                        eventobj.eventcontent.payload.messagepayload[
                                            eventobj.eventcontent.payload.messagepayload.index(
                                                self.componentinstancenumber) - 1],
                                        sequencenumber=eventobj.eventcontent.header.sequencenumber)
            self.send_down(Event(self, EventTypes.MFRT, GenericMessage(header, eventobj.eventcontent.payload)))

    def is_RDRequest_recevied_first_time(self, sequence_number):
        if sequence_number in self.RDRequestCache.keys():
            return False
        else:
            return True

    def RDRequest_received_time_check(self, sequence_number):
        if self.RDRequests[sequence_number] < N:
            self.RDRequests[sequence_number] += 1
            return True
        else:
            return False

    def check_if_in_RDrequest(self, evenyobj: Event):
        if self.componentinstancenumber in evenyobj.eventcontent.payload.messagepayload:
            return True
        else:
            return False



from ahc.Ahc import ComponentModel, Event, framers, EventTypes
from enum import Enum
import queue


class GenericMacEventTypes(Enum):
    HANDLEMACFRAME = "handlemacframe"


class GenericMac(ComponentModel):

    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)
        self.framequeue = queue.Queue()
        self.ahcuhd = framers.get_ahcuhd_by_id(self.componentinstancenumber)
        print("I am Generic MAC my uhd instance id is ", self.ahcuhd.componentinstancenumber)
        self.eventhandlers[GenericMacEventTypes.HANDLEMACFRAME] = self.on_handlemacframe
    
    def on_init(self, eventobj: Event):
        self.send_self(Event(self, GenericMacEventTypes.HANDLEMACFRAME, None))  # Continuously trigger handle_frame
        # print("Initialized", self.componentname, ":", self.componentinstancenumber)

    def on_handlemacframe(self, eventobj: Event): 
        # print("will handle frame from on_handlemacframe")     
        self.handle_frame()
        # self.send_self(Event(self, GenericMacEventTypes.HANDLEMACFRAME, None)) #Continuously trigger handle_frame
        # HANDLEMACFRAME event will be generated by the inheriting component to facilitate delay
    
    def on_message_from_bottom(self, eventobj: Event):
        # print(f"I am {self.componentname}, eventcontent={eventobj.eventcontent}\n")
        evt = Event(self, EventTypes.MFRB, eventobj.eventcontent)
        self.send_up(evt)
    
    def on_message_from_top(self, eventobj: Event):
        # print(f"I am {self.componentname}, eventcontent={eventobj.eventcontent}\n")
        # put message in queue and try accessing the channel
        self.framequeue.put_nowait(eventobj)
        #print("Mac put the frame in queueu", eventobj.eventcontent.payload)



from ahc.Ahc import ComponentModel, FramerObjects
from ahc.EttusUsrp.UhdUtils import AhcUhdUtils 
from ahc.EttusUsrp.LiquidDspUtils import *
from enum import Enum
from ahc.Ahc import Event, EventTypes, GenericMessage, GenericMessageHeader, GenericMessagePayload,MessageDestinationIdentifiers
from ctypes import *
import pickle

framers = FramerObjects()


# define your own message types
class UsrpB210PhyMessageTypes(Enum):
  PHYFRAMEDATA = "PHYFRAMEDATA"


class UsrpB210PhyEventTypes(Enum):
  RECV = "recv"


# define your own message header structure
class UsrpB210PhyMessageHeader(GenericMessageHeader):
    pass


# define your own message payload structure
class UsrpB210PhyMessagePayload(GenericMessagePayload):
    
  def __init__(self, header, payload):
    self.phyheader = header
    self.phypayload = payload


class FrameHandlerBase(ComponentModel):

    def __init__(self,componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)
        
        self.chan = 0
        self.bandwidth = 250000
        self.freq = 2462000000.0
        self.lo_offset = 0
        self.rate = self.bandwidth
        self.hw_tx_gain = 70.0  # hardware tx antenna gain
        self.hw_rx_gain = 20.0  # hardware rx antenna gain
        self.sw_tx_gain = -12.0  # software gain
        self.duration = 1
        self.ahcuhd = AhcUhdUtils(self.componentinstancenumber)
        framers.add_framer(id(self), self)
        framers.add_ahcuhd(componentinstancenumber, self.ahcuhd )
        self.ahcuhd.configureUsrp("winslab_b210_" + str(self.componentinstancenumber))        
        print("Configuring", "winslab_b210_" + str(self.componentinstancenumber))
        self.configure()
        self.eventhandlers[UsrpB210PhyEventTypes.RECV] = self.on_recv
        
    def on_recv(self, eventobj: Event):
        #print("Node", self.componentinstancenumber, " Received message type:", eventobj.eventcontent.header.messagetype, "  from ", eventobj.eventcontent.payload.phyheader.messagefrom)

        if eventobj.eventcontent.payload.phyheader.messagefrom != self.componentinstancenumber:
          msg = GenericMessage(eventobj.eventcontent.payload.phyheader, eventobj.eventcontent.payload.phypayload)
          self.send_up(Event(self, EventTypes.MFRB, msg))

        
    def on_message_from_top(self, eventobj: Event):
    # channel receives the input message and will process the message by the process event in the next pipeline stage
    # Preserve the event id through the pipeline
        
        str_header = "12345678"  #This is the PMD flexframe header. Ourt physical layer header will be concat with the payload below...
        hlen = len(str_header)
        byte_arr_header = bytearray(str_header, 'utf-8')
        header = (c_ubyte * hlen)(*(byte_arr_header))
        
        hdr = UsrpB210PhyMessageHeader(UsrpB210PhyMessageTypes.PHYFRAMEDATA, self.componentinstancenumber,MessageDestinationIdentifiers.LINKLAYERBROADCAST)
        pld = UsrpB210PhyMessagePayload(eventobj.eventcontent.header, eventobj.eventcontent.payload )
        msg = GenericMessage(hdr, pld)
        byte_arr_msg = bytearray(pickle.dumps(msg))
        plen = len(byte_arr_msg)
        payload = (c_ubyte * plen)(*(byte_arr_msg))
        payload_len = plen
        #print("bytearry:", byte_arr_msg, "Payload:",payload, " payload_len:", payload_len)
        self.transmit(header, payload, payload_len, LIQUID_MODEM_QPSK, LIQUID_FEC_NONE, LIQUID_FEC_HAMMING74 )  # TODO: Check params
        #print("sentpload=", string_at(payload, payload_len))
        #pload = string_at(payload, payload_len)
        #print("pload=", pload)
        #phymsg = pickle.loads(pload)
        #msg2 = GenericMessage(phymsg.header, phymsg.payload)
        



# https://www.etsi.org/deliver/etsi_en/300300_300399/300328/02.01.01_60/en_300328v020101p.pdf
#1) Before transmission, the equipment shall perform a Clear Channel Assessment (CCA) check using energy detect. The equipment shall observe the operating channel for the duration of the CCA observation time which shall be not less than 18 μs. The channel shall be considered occupied if the energy level in the channel exceeds the threshold given in step 5 below. If the equipment finds the channel to be clear, it may transmit immediately. See figure 2 below.
#3) The total time during which an equipment has transmissions on a given channel without re-evaluating the availability of that channel, is defined as the Channel Occupancy Time.
# The Channel Occupancy Time shall be in the range 1 ms to 10 ms followed by an Idle Period of at least 5 % of the Channel Occupancy Time used in the equipment for the current Fixed Frame Period.
# The energy detection threshold for the CCA shall be proportional to the transmit power of the transmitter: for a 20 dBm e.i.r.p. transmitter the CCA threshold level (TL) shall be equal to or less than -70 dBm/MHz at the input to the receiver assuming a 0 dBi (receive) antenna assembly. This threshold level (TL) may be corrected for the (receive) antenna assembly gain (G); however, beamforming gain (Y) shall not be taken into account. For power levels less than 20 dBm e.i.r.p. the CCA threshold level may be relaxed to:
#TL = -70 dBm/MHz + 10 × log10 (100 mW / Pout) (Pout in mW e.i.r.p.)

import uhd
import math
from threading import Thread, Lock
import numpy as np
import inspect


class AhcUhdUtils:
    INIT_DELAY = 0.05  # 50mS initial delay before transmit
    samps_per_est = 100
    #bandwidth = 250000
    #freq =2462000000.0
    lo_offset = 0
    
    wave_freq=10000
    wave_ampl = 0.3
    #hw_tx_gain = 70.0           # hardware tx antenna gain
    #hw_rx_gain = 20.0           # hardware rx antenna gain
    duration = 1
    
    
    def __init__(self, componentinstancenumber):
        self.mutex = Lock()
        self.cca = False
        self.componentinstancenumber = componentinstancenumber
        
    def on_init(self):
        pass
    
    def configureUsrp(self, devicename, type="b200", freq =2162000000.0, bandwidth = 1000000, chan = 0, hw_tx_gain = 50.0, hw_rx_gain = 20.0):
            
        self.devicename = devicename
        self.freq = freq
        self.bandwidth = bandwidth
        self.chan = chan
        self.hw_tx_gain = hw_tx_gain
        self.hw_rx_gain = hw_rx_gain
        self.tx_rate= self.bandwidth
        self.rx_rate= self.bandwidth
        print(f"Configuring type={type},devicename={devicename}, freq={freq}, bandwidth={bandwidth}, channel={chan}, hw_tx_gain={hw_tx_gain}, hw_rx_gain={hw_rx_gain}")
        self.usrp = uhd.usrp.MultiUSRP(f"name={devicename}")
        
        self.usrp.set_rx_bandwidth(self.bandwidth, self.chan)
        self.usrp.set_tx_bandwidth(self.bandwidth, self.chan)
        
        self.usrp.set_rx_freq(self.freq, self.chan)
        self.usrp.set_tx_freq(self.freq, self.chan)
        
        self.usrp.set_rx_bandwidth(self.bandwidth,self.chan)
        self.usrp.set_tx_bandwidth(self.bandwidth,self.chan)
        
        self.usrp.set_rx_rate(self.tx_rate, self.chan)
        self.usrp.set_tx_rate(self.rx_rate, self.chan)
        
        self.usrp.set_rx_gain(self.hw_rx_gain, self.chan)
        self.usrp.set_tx_gain(self.hw_tx_gain, self.chan)

        #self.usrp.set_rx_agc(True, self.chan)
        

        stream_args = uhd.usrp.StreamArgs('fc32', 'sc16')
        stream_args.channels = [self.chan]
        
        self.rx_streamer = self.usrp.get_rx_stream(stream_args)
        self.tx_streamer = self.usrp.get_tx_stream(stream_args)
    
    def get_usrp_power(self,num_samps=1000000, chan=0):
        uhd.dsp.signals.get_usrp_power(self.rx_streamer, num_samps, chan)
        
    
    def ischannelclear(self, threshold=-70, pout=100):
        self.cca = True
        cca_threshold = threshold + 10*math.log10(100/pout)
        tx_rate = self.usrp.get_rx_rate(self.chan) / 1e6
        samps_per_est = math.floor(18 * tx_rate)
        #samps_per_est = 10
        power_dbfs = 0
        self.mutex.acquire(1)
        try:
            power_dbfs = uhd.dsp.signals.get_usrp_power(self.rx_streamer, num_samps=int(samps_per_est), chan=self.chan)
        except Exception as e:
            print("Exception in CCA: ", e)
        finally:
            self.mutex.release()
            self.start_usrp_rx()
        #print("Power-dbfs=", power_dbfs)
        self.cca = False
        if (power_dbfs > cca_threshold ):
            #print(power_dbfs)
            return False, power_dbfs
        else:
            return True, power_dbfs
    
    def start_rx(self, rx_callback, framer):
        print(f"start_rx on usrp winslab_b210_{self.componentinstancenumber}")
        self.framer = framer
        self.rx_callback = rx_callback
        self.rx_rate = self.usrp.get_rx_rate()
        self.start_usrp_rx()
        t = Thread(target=self.rx_thread, args=[])
        t.daemon = True
        t.start()
        

    def start_usrp_rx(self):
        #print(f"start_usrp_rx on usrp winslab_b210_{self.componentinstancenumber}")
        stream_cmd = uhd.types.StreamCMD(uhd.types.StreamMode.start_cont)
        self.rx_streamer.issue_stream_cmd(stream_cmd)
        
    def stop_usrp_rx(self):
        self.rx_streamer.issue_stream_cmd(uhd.types.StreamCMD(uhd.types.StreamMode.stop_cont))
        
    def rx_thread(self):
        print(f"rx_thread on usrp winslab_b210_{self.componentinstancenumber}")
        #print(f"max_samps_per_packet={max_samps_per_packet}")
        
        #print(f"recv_buffer={recv_buffer")
        while(True):
            if self.cca == False:
                self.mutex.acquire(1)
                #print(f"rx_thread on usrp winslab_b210_{self.componentinstancenumber} ---> {self.devicename}")
                try:
                    had_an_overflow = False
                    rx_metadata = uhd.types.RXMetadata()
                    max_samps_per_packet = self.rx_streamer.get_max_num_samps()
                    recv_buffer = np.zeros( max_samps_per_packet, dtype=np.complex64)
                    
                    num_rx_samps = self.rx_streamer.recv(recv_buffer, rx_metadata)
                    
                    #print(f"num_rx_samps={num_rx_samps}")
                    self.rx_callback(num_rx_samps, recv_buffer)
                    #print(f"rx_thread on usrp winslab_b210_{self.componentinstancenumber} ---> {self.devicename}\t{self.rx_callback}")
                except RuntimeError as ex:
                    print("Runtime error in receive: %s", ex)
                finally:
                    self.mutex.release()
                if rx_metadata.error_code == uhd.types.RXMetadataErrorCode.none:
                    pass
                elif rx_metadata.error_code == uhd.types.RXMetadataErrorCode.overflow:
                    #print("Receiver error: overflow  %s, continuing...", rx_metadata.strerror())
                    pass
                elif rx_metadata.error_code == uhd.types.RXMetadataErrorCode.late:
                    #print("Receiver error: late %s, continuing...", rx_metadata.strerror())
                    pass
                elif rx_metadata.error_code == uhd.types.RXMetadataErrorCode.timeout:
                    print("Receiver error:timeout  %s, continuing...", rx_metadata.strerror())
                    pass
                else:
                    print("Receiver error: %s", rx_metadata.strerror())
                    
        
    def finalize_transmit_samples(self):   
        tx_metadata = uhd.types.TXMetadata() 
        tx_metadata.end_of_burst = True
        tx_metadata.start_of_burst = False
        tx_metadata.has_time_spec = False
        num_tx_samps = self.tx_streamer.send(np.zeros((1, 0), dtype=np.complex64), tx_metadata)
        #self.tx_streamer.send(np.zeros(1, dtype=np.complex64), tx_metadata)
        #self.start_usrp_rx()
        
    def transmit_samples(self, transmit_buffer):
        #self.stop_usrp_rx()
        tx_metadata = uhd.types.TXMetadata()
        tx_metadata.has_time_spec = False
        tx_metadata.start_of_burst = False
        tx_metadata.end_of_burst = False
        #print(transmit_buffer)
        num_tx_samps = self.tx_streamer.send(transmit_buffer, tx_metadata)
        #print("num_tx_samples", num_tx_samps)
        # Send a mini EOB packet
        #tx_metadata.end_of_burst = True
        #self.tx_streamer.send(np.zeros((1,0), dtype=np.complex), tx_metadata)
        


#!/usr/bin/env python

__author__ = "One solo developer"
__authors__ = ["Mahmoud Alasmar"]
__contact__ = "mahmoud.asmar@metu.edu.tr"
__date__ = "2021/05/26"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"



from enum import Enum
from ahc.Ahc import ComponentModel, Event
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Ahc import Topology


source = 0

class ApplicationLayerMessageTypes(Enum):
  START = "START"
  TOKEN = "TOKEN"
  VISITED = "VISITED"

class NodeState(Enum):
  IDLE = "IDLE"
  DISCOVERED = "DISCOVERED"

class NodeMark(Enum):
  visited = "visited"
  unvisited = "unvisited"
  father = "father"
  son = "son"

topo = Topology()

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
  pass

class ApplicationLayerComponent_Cidon(ComponentModel):
  #mark = {}
  #NeighbourList = {}

  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    self.NeighbourList = topo.get_neighbors(self.componentinstancenumber)
    self.state = NodeState.IDLE
    self.mark = {}
    for i in self.NeighbourList:
      self.mark[i] = NodeMark.unvisited
    self.numMesg  = 0

    if self.componentinstancenumber == source:
      destination = self.componentinstancenumber
      hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.START, self.componentinstancenumber,
                                          destination)
      payload = ApplicationLayerMessagePayload("23")
      proposalmessage = GenericMessage(hdr, payload)
      self.send_self(Event(self, "start", proposalmessage))
    else:
      pass

  def on_message_from_bottom(self, eventobj: Event):
    try:
      applmessage = eventobj.eventcontent
      hdr = applmessage.header
      if hdr.messagetype == ApplicationLayerMessageTypes.START:
        #print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "start", applmessage))
      elif hdr.messagetype == ApplicationLayerMessageTypes.TOKEN:
        #print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "token", applmessage))
      elif hdr.messagetype == ApplicationLayerMessageTypes.VISITED:
        #print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "visited", applmessage))
    except AttributeError:
      print("Attribute Error")

  def on_start(self, eventobj: Event):
    if self.state == NodeState.IDLE:
      self.state = NodeState.DISCOVERED
      self.Search()
      for i in self.NeighbourList:
        if self.mark[i] == NodeMark.visited or self.mark[i] == NodeMark.unvisited:
          hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.VISITED, self.componentinstancenumber,
                                                  i)
          payload = ApplicationLayerMessagePayload("23")
          proposalmessage = GenericMessage(hdr_new, payload)
          self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
          self.numMesg += 1


  def on_token(self, eventobj: Event):
    applmessage = eventobj.eventcontent
    hdr = applmessage.header
    source = hdr.messagefrom
    if self.state == NodeState.IDLE:
      self.mark[source] = NodeMark.father
      self.state = NodeState.DISCOVERED
      self.Search()
      for i in self.NeighbourList:
        if self.mark[i] == NodeMark.visited or self.mark[i] == NodeMark.unvisited:
          hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.VISITED, self.componentinstancenumber,
                                                  i)
          payload = ApplicationLayerMessagePayload("23")
          proposalmessage = GenericMessage(hdr_new, payload)
          self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
          self.numMesg += 1
    else:
      if self.mark[source] == NodeMark.unvisited:
        self.mark[source] = NodeMark.visited
      elif self.mark[source] == NodeMark.son:
        self.Search()


  def on_visited(self, eventobj: Event):
    applmessage = eventobj.eventcontent
    hdr = applmessage.header
    source = hdr.messagefrom
    if self.mark[source] == NodeMark.unvisited:
      self.mark[source] = NodeMark.visited
    elif self.mark[source] == NodeMark.son:
      self.mark[source] = NodeMark.visited
      self.Search()

  def Search(self):
    for i in self.NeighbourList:
      if self.mark[i] == NodeMark.unvisited:
        hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.TOKEN, self.componentinstancenumber,
                                                i)
        payload = ApplicationLayerMessagePayload("23")
        proposalmessage = GenericMessage(hdr_new, payload)
        self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
        self.numMesg += 1
        self.mark[i] = NodeMark.son
        return
      else:
        pass
    if self.componentinstancenumber == source:
      print(f"I am Node-{self.componentinstancenumber} local number messages sent is {self.numMesg}")
      print(f"Node-{self.componentinstancenumber} says all nodes were discovered, algorithm is finished")
      pass #  terminate application
    else:
      for i in self.NeighbourList:
        if self.mark[i] == NodeMark.father:
          hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.TOKEN, self.componentinstancenumber,
                                                  i)
          payload = ApplicationLayerMessagePayload("23")
          proposalmessage = GenericMessage(hdr_new, payload)
          self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
          self.numMesg += 1
          print(f"I am Node-{self.componentinstancenumber} local number messages sent is {self.numMesg}")
          return

  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers["start"] = self.on_start
    self.eventhandlers["token"] = self.on_token
    self.eventhandlers["visited"] = self.on_visited





import random
import time
from ahc.Ahc import ComponentModel, Event, EventTypes
import threading
from ahc.Routing.OLSR.OLSRSimplifiedComponent import OLSRMessageTypes
# TODO Not a proper implementations

class OLSRControlThread(threading.Thread):
    def __init__(self, event, component):
        threading.Thread.__init__(self)
        self.stopped = event
        self.increment = 0
        self.component = component
        self.wakeup = True

    def run(self):
        if self.wakeup is True:
            time.sleep(2)
        while not self.stopped.wait(1):
            self.wakeup = False
            self.increment += 1
            if self.increment == 2:
                self.component.generate_hello()

            if self.increment == 4:
                self.component.generate_hello()
                self.component.generate_tc()
                self.component.update_routing_table()
                self.increment = 0


class OLSRControlComponent(ComponentModel):

    def __init__(self, componentname, componentid):
        super(OLSRControlComponent, self).__init__(componentname, componentid)
        self.timer = threading.Event()
        self.control_thread = OLSRControlThread(self.timer, self).start()
        self.send_self(Event(self, EventTypes.INIT, None))

    def on_init(self, eventobj: Event):
        super(OLSRControlComponent, self).on_init(eventobj)
        # self.control_thread.start()

    def generate_hello(self):
        self.send_peer(Event(self, EventTypes.MFRP, OLSRMessageTypes.GENERATE_HELLO))

    def generate_tc(self):
        self.send_peer(Event(self, EventTypes.MFRP, OLSRMessageTypes.GENERATE_TC))

    def update_routing_table(self):
        self.send_peer(Event(self, EventTypes.MFRP, OLSRMessageTypes.UPDATE_TABLE))



import random 
import time 
from datetime import datetime
from enum import Enum 


import matplotlib.pyplot as plt
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout, to_agraph

from ahc.Ahc import ComponentModel, Event, GenericMessageHeader, GenericMessagePayload, GenericMessage, Topology, \
  MessageDestinationIdentifiers, EventTypes, Topology


class LogicalClockEventTypes(Enum): 
  INTERNAL = "internal"
  SEND = "senddown"

class LogicalClockMessageTypes(Enum):
  LCM = "LogicalClockMessage"

class LogicalClockMessageHeader(GenericMessageHeader):
  pass 


class LogicalClockMessagePayload(GenericMessagePayload):
  pass 


class LogicalClock(ComponentModel): 
  def on_init(self, eventobj: Event): 
    self.counter = 0 

    if self.componentinstancenumber == 0: 
      self.send_self(Event(self, EventTypes.MFRT, None))

  #this should serve as the on_send  
  def on_message_from_top(self, eventobj: Event): 
    randomnumber = random.randint(0,1)

    if randomnumber == 0: 
      #perfrom an internalEvent
      time.sleep(random.randint(0, 2))
      self.send_self((Event(self, LogicalClockEventTypes.INTERNAL, None)))
    else: 
      #send it to a random neighbour
      self.send_self(Event(self, LogicalClockEventTypes.SEND, None))

  def senddown(self, eventobj: Event): 
      #increment the counter before sending
      self.counter +=1
      neighbors = Topology().get_neighbors(self.componentinstancenumber)
      destination = neighbors[random.randint(0, len(neighbors)-1)]
      hdr = LogicalClockMessageHeader(LogicalClockMessageTypes.LCM, self.componentinstancenumber, destination)
      payload = LogicalClockMessagePayload(self.counter)
      message = GenericMessage(hdr, payload)
      time.sleep(random.randint(0, 2))
      self.send_down(Event(self, EventTypes.MFRT, message))

      print(f"Messaage sent from node {self.local_time(self.counter)} to node {destination}\n")


  #this should serve as the on_receive
  def on_message_from_bottom(self, eventobj: Event): 
    #after receiving a message you should increment the counter 
    #do some demultiplexing 
    msg = eventobj.eventcontent
    hdr = msg.header 
    payload = msg.payload
    incomingtimestamp = payload.messagepayload
    #update the timer of the payload 
    payload = incomingtimestamp
    self.counter = self.update_timestamp(incomingtimestamp, self.counter)

    print(f"Messaage received at node {self.local_time(self.counter)} from node {hdr.messagefrom}\n")

    self.send_up(Event(self, EventTypes.MFRB, payload))


  def on_internal(self, eventobj: Event): 
    #Increment counter so that we know something happened
    self.counter +=1 
    print(f"Internal event at node {self.local_time(self.counter)}\n")
    randomnumber = random.randint(0,1)
    if randomnumber == 0: 
      self.send_self(Event(self, LogicalClockEventTypes.INTERNAL, None))
    else: 
      self.send_self(Event(self, LogicalClockEventTypes.SEND, None))


  def local_time(self, counter): 
     return f"{self.componentinstancenumber} LOGICAL_TIME = {counter}"


  def update_timestamp(self, recv_time_stamp, counter):
    counter = max(recv_time_stamp, counter)+1
    return counter 


  def __init__(self, componentname, componentinstancenumber): 
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers[LogicalClockEventTypes.INTERNAL] = self.on_internal
    self.eventhandlers[LogicalClockEventTypes.SEND] = self.senddown


#IMPLEMENTATION OF VECTOR CLOCKS
class VectorClock(LogicalClock):
  def on_init(self, eventobj: Event): 
    N = len(Topology().nodes)
    self.counter = [0] * N

    if self.componentinstancenumber == 0: 
      self.send_self(Event(self, EventTypes.MFRT, None))
    else: 
      pass


  def on_message_from_top(self, eventobj: Event): 
    randomnumber = random.randint(0, 1)

    if randomnumber == 0: 
      time.sleep(random.randint(0, 2))
      self.send_self(Event(self, LogicalClockEventTypes.INTERNAL, None))
    elif randomnumber ==1: 
      self.send_self(Event(self, LogicalClockEventTypes.SEND, None))
  

  def senddown(self, eventobj: Event): 
    self.update_counter()
    neighbors = Topology().get_neighbors(self.componentinstancenumber)
    destination = neighbors[random.randint(0, len(neighbors)-1)]
    hdr = LogicalClockMessageHeader(LogicalClockMessageTypes.LCM, self.componentinstancenumber, destination)
    messagepayload = self.counter[:]
    payload = LogicalClockMessagePayload(messagepayload)
    message = GenericMessage(hdr, payload)
    time.sleep(random.randint(0,2))
    self.send_down(Event(self, EventTypes.MFRT, message))

    print(f"Messaage sent from node {self.local_time(self.counter)} to node {destination}\n") 


  def on_message_from_bottom(self, eventobj: Event):
    msg = eventobj.eventcontent
    hdr = msg.header
    payload = msg.payload
    incomingtimestamps = payload.messagepayload

    #Adjust your timestamps
    self.update_counter()
    self.update_timestamp(incomingtimestamps, self.counter)
    self.send_up(Event(self, EventTypes.MFRB, None))

    print(f"Messaage received at node {self.local_time(self.counter)} from node {hdr.messagefrom}\n")


  def on_internal(self, eventobj: Event):
    self.update_counter()
    print(f"Internal event at node {self.local_time(self.counter)}\n")

    randomnumber = random.randint(0,1)
    if randomnumber == 0: 
      time.sleep(random.randint(0,2))
      self.send_self(Event(self, LogicalClockEventTypes.INTERNAL, None))
    else: 
      self.send_self(Event(self, LogicalClockEventTypes.SEND, None))

    #randomly decide whether to haev an internal or senddown event
    self.send_self(Event(self, LogicalClockEventTypes.SEND, None))


  def update_timestamp(self, incoming_time_stamp, my_time_stamp): 
    for i in range(len(my_time_stamp)): 
      my_time_stamp[i] = max(incoming_time_stamp[i], my_time_stamp[i])
  
    self.counter = my_time_stamp


  def update_counter(self): 
    self.counter[self.componentinstancenumber] +=1


  def __init__(self, componentname, componentinstancenumber): 
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers[LogicalClockEventTypes.INTERNAL] = self.on_internal
    self.eventhandlers[LogicalClockEventTypes.SEND] = self.senddown
    self.counter = []














from enum import Enum

from ahc.Ahc import ComponentModel, Event, GenericMessageHeader, GenericMessagePayload, GenericMessage, Topology, \
  MessageDestinationIdentifiers, EventTypes

# define your own message types
class NetworkLayerMessageTypes(Enum):
  NETMSG = "NETMSG"

# define your own message header structure
class NetworkLayerMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class NetworkLayerMessagePayload(GenericMessagePayload):
  pass

class AllSeingEyeNetworkLayer(ComponentModel):

  def on_message_from_top(self, eventobj: Event):
    # Encapsulate the SDU in network layer PDU
    applmsg = eventobj.eventcontent
    destination = applmsg.header.messageto
    nexthop = Topology().get_next_hop(self.componentinstancenumber, destination)
    if nexthop != float('inf'):
      # print(f"{self.componentinstancenumber} will SEND a message to {destination} over {nexthop}")
      hdr = NetworkLayerMessageHeader(NetworkLayerMessageTypes.NETMSG, self.componentinstancenumber, destination,
                                      nexthop)
      payload = eventobj.eventcontent
      msg = GenericMessage(hdr, payload)
      self.send_down(Event(self, EventTypes.MFRT, msg))
    else:
      pass
      # print(f"NO PATH: {self.componentinstancenumber} will NOTSEND a message to {destination} over {nexthop}")

  def on_message_from_bottom(self, eventobj: Event):
    msg = eventobj.eventcontent
    hdr = msg.header
    payload = msg.payload

    if hdr.messageto == self.componentinstancenumber or hdr.messageto == MessageDestinationIdentifiers.NETWORKLAYERBROADCAST:  # Add if broadcast....
      self.send_up(Event(self, EventTypes.MFRB, payload))
      # print(f"I received a message to {hdr.messageto} and I am {self.componentinstancenumber}")
    else:
      destination = hdr.messageto
      nexthop = Topology().get_next_hop(self.componentinstancenumber, destination)
      if nexthop != float('inf'):
        newhdr = NetworkLayerMessageHeader(NetworkLayerMessageTypes.NETMSG, self.componentinstancenumber, destination,
                                           nexthop)
        newpayload = eventobj.eventcontent.payload
        msg = GenericMessage(newhdr, newpayload)
        self.send_down(Event(self, EventTypes.MFRT, msg))
        # print(f"{self.componentinstancenumber} will FORWARD a message to {destination} over {nexthop}")
      else:
        pass
        # print(f"NO PATH {self.componentinstancenumber} will NOT FORWARD a message to {destination} over {nexthop}")

  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)



import os
import sys
import time
import random
from enum import Enum

sys.path.insert(0, os.getcwd())

import networkx as nx
import matplotlib.pyplot as plt

from ahc.Ahc import ComponentModel, Event, ConnectorTypes, Lock, Thread, Topology, MessageDestinationIdentifiers
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Channels.Channels import P2PFIFOPerfectChannel
from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer

registry = ComponentRegistry()

# define your own message types
class ApplicationLayerMessageTypes(Enum):
    HELLO_MESSAGE = "HELLO_MESSAGE"
    USER_MESSAGE = "USER_MESSAGE"
    RREQ_MESSAGE = "RREQ_MESSAGE"
    RREP_MESSAGE = "RREP_MESSAGE"
    RERR_MESSAGE = "RERR_MESSAGE"

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
    def __init__(self, messagetype, messagefrom, messageto, nexthop=float('inf'), interfaceid=float('inf'), sequencenumber=-1):
        self.messagetype = messagetype
        self.messagefrom = messagefrom
        self.messageto = messageto
        self.nexthop = nexthop
        self.sender = messagefrom
        self.hopcount = 0
        self.broadcastid = -1
        self.sourcesequencenumber = -1
        self.destinationsequencenumber = -1
        self.feasibledistance = -1 
        self.Tbit = 0
        super(ApplicationLayerMessageHeader, self).__init__(messagetype, messagefrom, messageto, nexthop=float('inf'), interfaceid=float('inf'), sequencenumber=-1)

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
    pass


class ApplicationLayerComponent(ComponentModel):

    #def __init__(self, componentname, componentinstancenumber, context=None, configurationparameters=None, num_worker_threads=1):
    #    super().__init__(componentname, componentinstancenumber, context=None, configurationparameters=None, num_worker_threads=1)

    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

        self.broadcastdb = []
        self.routing_table = dict()

        #for testing purpose, initiate some messages
        if self.componentinstancenumber == 0:
            # destination = random.randint(len(Topology.G.nodes))
            destination = 1
            hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.HELLO_MESSAGE, self.componentinstancenumber,
                                                destination, nexthop=destination)
            payload = ApplicationLayerMessagePayload("23")
            hellomessage = GenericMessage(hdr, payload)
            randdelay = random.randint(0, 5)
            time.sleep(randdelay)
            self.send_self(Event(self, "HELLO_MESSAGE", hellomessage))

        elif self.componentinstancenumber == 2:
            destination = 0
            hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.RREQ_MESSAGE, self.componentinstancenumber,
                                             destination)
            payload = ApplicationLayerMessagePayload("23")          
            hdr.messagetype = ApplicationLayerMessageTypes.RREQ_MESSAGE
            hdr.messagefrom = self.componentinstancenumber
            hdr.nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST
            hdr.messageto = destination
            hdr.sender = self.componentinstancenumber
            hdr.hopcount = 0
            hdr.broadcastid = 0
            hdr.sourcesequencenumber = -1
            hdr.destinationsequencenumber = ""
            hdr.feasibledistance = 0            
            randdelay = random.randint(0, 5)
            time.sleep(randdelay)
            rreqmessage = GenericMessage(hdr, payload)
            self.send_self(Event(self, "RREQ_MESSAGE", rreqmessage))

    def on_message_from_bottom(self, eventobj: Event):
        try:
            applmessage = eventobj.eventcontent
            hdr = applmessage.header
            payload = applmessage.payload

            #self.queue_lock.acquire() # protect message_queue, both component thread and Toueg thread are trying to access data
            #self.message_queue.append((int(hdr.messagefrom), hdr.message_type, payload))
            #self.queue_lock.release()


            if hdr.messagetype == ApplicationLayerMessageTypes.HELLO_MESSAGE:
               print(
                f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")

            elif hdr.messagetype == ApplicationLayerMessageTypes.USER_MESSAGE:
                print(
                    f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")

            elif hdr.messagetype == ApplicationLayerMessageTypes.RREQ_MESSAGE:
                print(
                    f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")

            elif hdr.messagetype == ApplicationLayerMessageTypes.RREP_MESSAGE:
                    print(
                    f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message") 

            elif hdr.messagetype == ApplicationLayerMessageTypes.RERR_MESSAGE:
                print(
                    f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")

        except AttributeError:
            print("Attribute Error")

    # print(f"{self.componentname}.{self.componentinstancenumber}: Gotton message {eventobj.content} ")
    # value = eventobj.content.value
    # value += 1
    # newmsg = MessageContent( value )
    # myevent = Event( self, "agree", newmsg )
    # self.trigger_event(myevent)

    def update_topology(self):
        Topology().nodecolors[self.componentinstancenumber] = 'r'
        Topology().plot()

    def __init__(self, componentname, componentinstancenumber): 
        super().__init__(componentname, componentinstancenumber)
        self.eventhandlers["HELLO_MESSAGE"] = self.on_HelloMessage
        self.eventhandlers["USER_MESSAGE"] = self.on_UserMessage
        self.eventhandlers["RREQ_MESSAGE"] = self.on_RreqMessage
        self.eventhandlers["RREP_MESSAGE"] = self.on_RrepMessage
        self.eventhandlers["RERR_MESSAGE"] = self.on_RerrMessage
        self.eventhandlers["timerexpired"] = self.on_timer_expired
        

    def on_HelloMessage(self, eventobj: Event):
        applmessage = eventobj.eventcontent
        hdr = applmessage.header
        payload = applmessage.payload
        if (hdr.sender not in self.routing_table.keys()):
            destination = hdr.sender
            hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.HELLO_MESSAGE, self.componentinstancenumber,
                                            destination, self.componentinstancenumber)
            payload = ApplicationLayerMessagePayload("23")
            hellomessage = GenericMessage(hdr, payload)
            self.send_down(Event(self, EventTypes.MFRT, hellomessage))

    def on_UserMessage(self, eventobj: Event):
        print(f"User Message: {eventobj.eventcontent}")

    def on_RreqMessage(self, eventobj: Event):
        print(f"RREQ Message: {eventobj.eventcontent}")
        applmessage = eventobj.eventcontent
        hdr = applmessage.header
        payload = applmessage.payload
        hdr.hopcount = hdr.hopcount + 1
        # Check if we have a route to the source. If we have, see if we need
        # to update it. Specifically, update it only if:
        #
        # 1. The destination sequence number for the route is less than the
        #    originator sequence number in the packet
        # 2. The sequence numbers are equal, but the hop_count in the packet
        #    + 1 is lesser than the one in routing table
        # 3. The sequence number in the routing table is unknown
        #
        # If we don't have a route for the originator, add an entry

        if (hdr.Tbit == 1):
            #forward rreq message immediately
            hdr.hopcount = hdr.hopcount + 1 
            hdr.sender = self.componentinstancenumber
            hdr.messageto = MessageDestinationIdentifiers.NETWORKLAYERBROADCAST
            hdr.nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST
            print(
                f"{self.componentinstancenumber} will forward the message to {hdr.messageto} over {hdr.nexthop}")
            broadcastmessage = GenericMessage(hdr, applmessage)
            
            self.send_down(Event(self, EventTypes.MFRT, broadcastmessage))
            if hdr.broadcastid not in self.broadcastdb:
                self.broadcastdb.append(hdr.broadcastid)
        
        else:

            if hdr.messagefrom in self.routing_table.keys():
               route = self.routing_table[hdr.messagefrom]
               if (hdr.feasibledistance < (int(route['feasibledistance']))):
                #start new rreq with T=1 from this node
                   hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.RREQ_MESSAGE, self.componentinstancenumber,
                                                    hdr.messageto)
                   payload = ApplicationLayerMessagePayload("23")

                   
                   hdr.messagetype = ApplicationLayerMessageTypes.RREQ_MESSAGE
                   hdr.messagefrom = hdr.messagefrom
                   hdr.nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST
                   hdr.messageto = hdr.messageto
                   hdr.sender = self.componentinstancenumber
                   hdr.hopcount = 0
                   hdr.broadcastid = hdr.broadcastid + 1
                   hdr.sourcesequencenumber =  hdr.sourcesequencenumber + 1
                   hdr.destinationsequencenumber = ""
                   hdr.feasibledistance = 0
                   hdr.Tbit = 1
                   
                   rreqmessage = GenericMessage(hdr, payload)
                  
                   self.send_down(Event(self, EventTypes.MFRT, rreqmessage))

               else:
                #send RREP to the source
                    route = self.routing_table[hdr.messagefrom]
                    if (int(route['sequencenumber']) < hdr.sourcesequencenumber):
                        route['sequencenumber'] = hdr.sourcesequencenumber

                    elif (int(route['sequencenumber']) == hdr.sourcesequencenumber):
                        if (int(route['hopcount']) > hdr.hopcount):
                            route['hopcount'] = hdr.hopcount
                            route['nexthop'] = hdr.sender

                    elif (int(route['sequencenumber']) == -1):
                        route['sequencenumber'] = hdr.sourcesequencenumber

                    hdr.messagetype == ApplicationLayerMessageTypes.RREP_MESSAGE
                    hdrmessageto = hdr.messagefrom
                    hdr.messagefrom = self.componentinstancenumber
                    hdr.nexthop = hdr.sender
                    hdr.sender = self.componentinstancenumber
                    hdr.sourcesequencenumber = ""
                    hdr.destinationsequencenumber = hdr.destinationsequencenumber + 1
                    hdr.feasibledistance = int(route['feasibledistance'])
                    hdr.hopcount = 0
                    
                    print(
                        f"{self.componentinstancenumber} will SEND RREP message to {hdr.messageto} over {hdr.nexthop}")
                    RREPmessage = GenericMessage(hdr, applmessage)
                    self.send_down(Event(self, EventTypes.MFRT, RREPmessage))

            else:
                self.routing_table[hdr.messagefrom] = {'Destination': str(hdr.messagefrom), 
                                    'nexthop': str(hdr.sender),
                                    'sequencenumber': str(hdr.sourcesequencenumber),
                                    'feasibledistance': str(hdr.hopcount),
                                    'hopcount': str(hdr.hopcount)}

            if hdr.messageto == self.componentinstancenumber:

               # Check if we are the destination. If we are, generate and send an
               # RREP back.
               #
                 hdr.messagetype == ApplicationLayerMessageTypes.RREP_MESSAGE
                 hdrmessageto = hdr.messagefrom
                 hdr.messagefrom = self.componentinstancenumber
                 hdr.nexthop = hdr.sender
                 hdr.sender = self.componentinstancenumber
                 hdr.sourcesequencenumber = ""
                 hdr.destinationsequencenumber = hdr.destinationsequencenumber + 1
                 hdr.hopcount = 0

                 print(
                     f"{self.componentinstancenumber} will SEND RREP message to {hdr.messageto} over {hdr.nexthop}")
                 RREPmessage = GenericMessage(hdr, applmessage)
                 self.send_down(Event(self, EventTypes.MFRT, RREPmessage))
               
             
            else:

                if hdr.broadcastid in self.broadcastdb:
                  pass  # we have already handled this flooded message
                else:
                  # Forwarding rreq, send to lower layers
                  self.update_topology()
                  hdr.hopcount = hdr.hopcount + 1 
                  hdr.sender = self.componentinstancenumber
                  hdr.messageto = MessageDestinationIdentifiers.NETWORKLAYERBROADCAST
                  hdr.nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST
                  print(
                      f"{self.componentinstancenumber} will forward the message to {hdr.messageto} over {hdr.nexthop}")
                  broadcastmessage = GenericMessage(hdr, applmessage)
                  
                  self.send_down(Event(self, EventTypes.MFRT, broadcastmessage))
                  self.broadcastdb.append(hdr.broadcastid)


    def on_RrepMessage(self, eventobj: Event):
        print(f"RREP Message: {eventobj.eventcontent}")
        applmessage = eventobj.eventcontent
        hdr = applmessage.header
        payload = applmessage.payload

        # We are not the destination. Check if we have a valid route
        # to the destination. If we have, generate and send back an
        # RREP.
        #
        if (hdr.messageto in self.routing_table.keys()):
           # Verify that the route is valid and has a higher seq number
           route = self.routing_table[hdr.messageto]
           route_dest_seq_no = int(route['sequencenumber'])
           if (route_dest_seq_no >= hdr.destinationsequencenumber):
              hdr.messagetype == ApplicationLayerMessageTypes.RREP_MESSAGE
              hdr.nexthop = route_dest_seq_no
              hdr.sender = self.componentinstancenumber
              hdr.destinationsequencenumber = route_dest_seq_no
              hdr.hopcount = int(route['hopcount'])

              #print(f"{self.componentinstancenumber} will forward RREP message to {destination} over {nexthop}")
              RREPmessage = GenericMessage(hdr, applmessage)
              self.send_down(Event(self, EventTypes.MFRT, RREPmessage))



    def on_RerrMessage(self, eventobj: Event):
        print(f"RERR Message: {eventobj.eventcontent}")

    def on_timer_expired(self, eventobj: Event):
        pass
        

class LDRnode(ComponentModel):

    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

        ## the first process does not start immediate, it stars with a peer message
        #if self.componentinstancenumber != 0:
        #    thread = Thread(target=self.job, args=[45, 54, 123])
        #    thread.start()

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def __init__(self, componentname, componentid):

        # SUBCOMPONENTS
        self.appllayer = ApplicationLayerComponent("ApplicationLayer", componentid)
        self.netlayer = AllSeingEyeNetworkLayer("NetworkLayer", componentid)
        self.linklayer = LinkLayer("LinkLayer", componentid)
        # self.failuredetect = GenericFailureDetector("FailureDetector", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.appllayer.connect_me_to_component(ConnectorTypes.DOWN, self.netlayer)
        # self.failuredetect.connectMeToComponent(PortNames.DOWN, self.netlayer)
        self.netlayer.connect_me_to_component(ConnectorTypes.UP, self.appllayer)
        # self.netlayer.connectMeToComponent(PortNames.UP, self.failuredetect)
        self.netlayer.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.netlayer)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        #self.message_queue = [] # for the next invication clear it...
        #self.queue_lock = Lock()

        super().__init__(componentname, componentid)


    #def job(self, *arg):
    #    self.all_process_ids = []
    #    for element in ComponentRegistry().components:
    #            self.all_process_ids.append(element)
    #    print("Available nodes : ", self.all_process_ids)
    #    self.neighbors = Topology().get_neighbors(self.componentinstancenumber) # retrieve all neighbor ids...





from random import choice
import time
from enum import Enum
from typing import Any, Dict, List
import uuid

from ahc.Ahc import ComponentModel, Event, GenericMessageHeader, GenericMessagePayload, GenericMessage, Topology, \
  MessageDestinationIdentifiers, EventTypes

# define your own message types
class DfsMessageTypes(Enum):
  FORWARD = "@tarrys/forward"
  START = "@tarrys/start"

# define your own message header structure
class DfsMessageHeader(GenericMessageHeader):
  def __init__(self, *args, token, **kwargs):
    super().__init__(*args, **kwargs)
    self.token = token

# define your own message payload structure
class DfsMessagePayload(GenericMessagePayload):
  pass

class DfsNeighbor:
  def __init__(self, id, invoked):
    self.id = id
    self.invoked = invoked

class DfsTraverse(ComponentModel):
  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.token_neighbor_mapping = {}
    self.token_parent_mapping = {}

  def on_message_from_bottom(self, eventobj: Event):
    msg = eventobj.eventcontent
    hdr = msg.header
    message_source = hdr.messagefrom

    payload:List[Any] = msg.payload.messagepayload

    if hdr.messagetype == DfsMessageTypes.FORWARD or hdr.messagetype == DfsMessageTypes.START:
      token = hdr.token
      if hdr.messagetype == DfsMessageTypes.START:
        self.token_parent_mapping[token] = -1
  
      parent_for_token = self.token_parent_mapping.get(token, None)
      if parent_for_token == None:
        self.token_parent_mapping[token] = message_source
        parent_for_token = message_source

      message = None
      next_target = None
      """
        DFS's algorithm has 3 rules.
        1 - A process never forwards the token through the same channel twice.
        2 - A process only forwards the token to its parent when there is no other option.
        3 - When a process receives the token, it immediately sends it back through the same 
            channel if this is allowed by rules 1 and 2.
      """
      uninvoked_neighbors = [n for n in self.get_neighbor_mapping_for_token(token) if n.invoked == False and n.id != parent_for_token and n.id != message_source]
      # Send message back if possible
      if message_source in [n.id for n in self.get_neighbor_mapping_for_token(token) if n.invoked == False] and message_source != parent_for_token:
          neigh = [n for n in self.get_neighbor_mapping_for_token(token) if n.invoked == False and n.id == message_source][0]
          neigh.invoked = True
          next_target = message_source
      elif len(uninvoked_neighbors) > 0: # If true, send to the available neighbor
        neigh = choice(uninvoked_neighbors)
        neigh.invoked = True
        next_target = neigh.id
      else: # Else, send the token back to the parent
        if parent_for_token == -1: # If I am the initiator, traversing is completed
          print(payload)
          print("->".join(payload))
          print("TRAVERSING IS COMPLETED IN " + str(len(payload)) + " hops")
          print(f"Graph had {Topology().G.number_of_edges()} edges")
          print(len(set(payload)))
          return
        else:
          next_target = parent_for_token

      payload.append(str(self.componentinstancenumber))
      message = self.prepare_message(DfsMessageTypes.FORWARD, next_target, token, payload)
      self.send_down(Event(self, EventTypes.MFRT, message))

  def start_traverse(self):
    token = self.create_token()
    self.send_self(Event(self, EventTypes.MFRB, self.prepare_message(DfsMessageTypes.START, self.componentinstancenumber, token, [])))

  def create_token(self):
    return str(uuid.uuid4())

  def prepare_neighbor_map(self):
    neighbor_list = Topology().get_neighbors(self.componentinstancenumber)
    return [DfsNeighbor(n, False) for n in neighbor_list]
  
  def get_neighbor_mapping_for_token(self, token: str) -> List[DfsNeighbor]:
    mapping = self.token_neighbor_mapping.get(token)
    if mapping == None:
      mapping = self.prepare_neighbor_map()
      self.token_neighbor_mapping[token] = mapping
    return mapping

  def prepare_message(self, message_type: DfsMessageTypes, neighbor: int, token: str, payload:Any = None) -> GenericMessage:
    header = DfsMessageHeader(message_type, self.componentinstancenumber, neighbor, neighbor, token=token)
    payload = DfsMessagePayload(payload)
    return GenericMessage(header, payload)


from enum import Enum
from ahc.Ahc import GenericMessageHeader

class AODVMessageTypes(Enum):
    RREQ = "RREQ"
    RREP = "RREP"
    PROPOSE = "PROPOSE"
    ACCEPT = "ACCEPT"

class AODVMessageHeader(GenericMessageHeader):
    _broadcastid = 1

    def __init__(self, messagetype, messagefrom, messageto, hopcount, nexthop=float('inf'), interfaceid=float('inf'), sequencenumber=1):
        super().__init__(messagetype,messagefrom, messageto, nexthop, interfaceid, sequencenumber)
        self.hopcount = hopcount
        self.broadcastid = AODVMessageHeader._broadcastid
        AODVMessageHeader._broadcastid += 1
        
        

        




import queue
import random
from enum import Enum
from datetime import datetime as dt

from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer
from ahc.Ahc import (ComponentModel, Event, ConnectorTypes, ComponentRegistry, GenericMessagePayload, GenericMessageHeader,
                 GenericMessage, EventTypes)

registry = ComponentRegistry()

# define your own message types
class ApplicationLayerMessageType(Enum):
    BASIC = "basic"
    CONTROL = "control" # NOTE: Means the acknowledgement message!

class DSAHCNodeSimulationStatus(Enum):
    ACTIVE = "active"
    PASSIVE = "passive"
    OUT_OF_CLOCK = "ooc"
    OUT_OF_TREE = "oot"

    def __str__(self):
        return self.name

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
    pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
    pass

class DijkstraScholtenApplicationLayerComponent(ComponentModel):
    def __init__(self, componentname, componentinstancenumber, context):
        super().__init__(componentname, componentinstancenumber, context=context)

        self.context = context
        # self.eventhandlers[ApplicationLayerMessageType.BASIC] = self.on_basic_message
        # self.eventhandlers[ApplicationLayerMessageType.CONTROL] = self.on_control_message

        self.basic_message_queue = queue.Queue(maxsize=-1)
        self.control_message_queue = queue.Queue(maxsize=-1)
        self.simulation_state = DSAHCNodeSimulationStatus.PASSIVE

        self.sleep_ms_per_tick = context["ms_per_tick"]
        self.simulation_ticks_total = context["simulation_ticks"]
        self.alive_for_next_ticks = context["initial_liveness"][componentinstancenumber]
        self.communication_on_active_prob = context["communication_on_active_prob"]
        self.min_activeness_after_receive = context["min_activeness_after_receive"]
        self.max_activeness_after_receive = context["max_activeness_after_receive"]
        self.package_process_per_tick = context["node_package_process_per_tick"]
        self.die_passiveness_threshold = context["passiveness_death_thresh"]

        if context["hard_stop_on_tick"] is None:
            self.hard_stop_on_tick = None
        else:
            self.hard_stop_on_tick = context["hard_stop_on_tick"][self.componentinstancenumber]

        self.__tick_n = 0
        self._passive_counter = 0

        self._child_counter = 0
        self._parent_node = None
        self._in_tree = False
        self._children = []
        
        self._cms = 0

        self._i_am_root = context["network"].root == self.componentinstancenumber

        if context["only_root_alive_initially"]:
            if self._i_am_root:
                self.alive_for_next_ticks = 20 # this may change though...
                self._in_tree = True
            else:
                self.alive_for_next_ticks = 0
        elif self._i_am_root and self.alive_for_next_ticks == 0:
            self.alive_for_next_ticks = 20 # this may change though...

        if self.alive_for_next_ticks > 0:
            self.simulation_state = DSAHCNodeSimulationStatus.ACTIVE

    def prepare_application_layer_message(self, message_type: ApplicationLayerMessageType, destination_node_id: int, payload: object) -> GenericMessage:
        hdr = ApplicationLayerMessageHeader(message_type, self.componentinstancenumber, destination_node_id)
        payload = ApplicationLayerMessagePayload(payload)
        
        return GenericMessage(hdr, payload)

    def send_random_basic_message(self, to: int) -> None:
        self._child_counter += 1
        self._children.append(to)
        self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.BASIC, to, str(dt.now().timestamp()))))

    def send_ack_control_message(self, to: int, is_dead: bool) -> None:
        # print(f"send_ack_control_message: N-{self.componentinstancenumber} ==> N-{to} ({self._parent_node}) : {'DEAD' if is_dead else 'PKG_RESP'}")
        self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.CONTROL, to, str(dt.now().timestamp()))))
        self._cms += 1

    def on_init(self, eventobj: Event):
        # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
        pass

    def on_message_from_bottom(self, eventobj: Event):
        applmessage = eventobj.eventcontent
        hdr = applmessage.header

        # print(f"Node-{self.componentinstancenumber}: Node-{hdr.messagefrom} has sent {hdr.messagetype} message (payload: {applmessage.payload})")

        if hdr.messagetype == ApplicationLayerMessageType.BASIC:
            self.basic_message_queue.put_nowait(applmessage)

            if self._in_tree:
                self.send_ack_control_message(hdr.messagefrom, False)
            else:
                self._parent_node = hdr.messagefrom
                self._in_tree = True

                if self.componentinstancenumber not in self.context["alive_nodes"]:
                    self.context["alive_nodes"].append(self.componentinstancenumber)

        elif hdr.messagetype == ApplicationLayerMessageType.CONTROL:
            # self.control_message_queue.put_nowait(applmessage)

            try:
                self._children.remove(hdr.messagefrom)
                self._child_counter -= 1
            except ValueError as e:
                # print(f"\n\n\n{self.componentinstancenumber}: {e} {hdr.messagefrom} {self._children}\n\n\n")
                pass

    def exit_tree(self):
        # Exit from the tree
        if self._in_tree:
            self._in_tree = False
            if self._parent_node is not None:
                self.send_ack_control_message(self._parent_node, True)
        
        self._parent_node = None

        if self.componentinstancenumber in self.context["alive_nodes"]:
            self.context["alive_nodes"].remove(self.componentinstancenumber)

    def simulation_tick(self):
        next_state = None
        got_packages_from = None
        to_friend = None

        _upd_children = []

        for c in self._children:
            if c in self.context["alive_nodes"]:
                _upd_children.append(c)

        self._children = _upd_children
        self._child_counter = len(self._children)

        if self.simulation_state == DSAHCNodeSimulationStatus.OUT_OF_TREE:
            next_state = DSAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: OOT")            #NOTE: DEV
            # print(f"   ==> N-{self.componentinstancenumber}: OUT OF TREE")    
        elif self.__tick_n >= self.simulation_ticks_total:
            self.exit_tree()
            next_state = DSAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: OOC DEAD")       #NOTE: DEV
        elif not self._i_am_root and self._passive_counter >= self.die_passiveness_threshold:
            self.exit_tree()
            next_state = DSAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: PASSIVE DIE")    #NOTE: DEV
        elif not self._i_am_root and self.__tick_n >= self.hard_stop_on_tick:
            self.exit_tree()
            next_state = DSAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: HARD STOP")      #NOTE: DEV
        else:
            if self.simulation_state == DSAHCNodeSimulationStatus.OUT_OF_CLOCK:
                next_state = DSAHCNodeSimulationStatus.OUT_OF_CLOCK
            elif self.simulation_state == DSAHCNodeSimulationStatus.PASSIVE:
                if self.basic_message_queue.empty():
                    if self._in_tree and self._child_counter == 0:
                        if self._i_am_root:
                            print(f"  **ROOT** N-{self.componentinstancenumber}: Termination!!!")
                            __cms =  self._cms
                            self._cms = 0
                            return None, None, __cms
                        else:
                            self.exit_tree()
                            next_state = DSAHCNodeSimulationStatus.PASSIVE
                            print(f"   ==> N-{self.componentinstancenumber}: OUT OF TREE / PASSIVE")
                    else:
                        # no incoming package, still passive.
                        next_state = DSAHCNodeSimulationStatus.PASSIVE
                else:
                    got_packages_from = []

                    for _ in range(self.package_process_per_tick):
                        try:
                            package = self.basic_message_queue.get_nowait()
                            # print(f"+P+ N-{self.componentinstancenumber} <==BASIC== N-{package.header.messagefrom} ({package.payload.messagepayload})")
                            got_packages_from.append(package.header.messagefrom)
                        except queue.Empty:
                            break

                    self.alive_for_next_ticks = random.randint(self.min_activeness_after_receive, self.max_activeness_after_receive)
                    next_state = DSAHCNodeSimulationStatus.ACTIVE
            elif self.simulation_state == DSAHCNodeSimulationStatus.ACTIVE:
                got_packages_from = []

                for _ in range(self.package_process_per_tick):
                    try:
                        package = self.basic_message_queue.get_nowait()
                        # print(f"+A+ N-{self.componentinstancenumber} <==BASIC== N-{package.header.messagefrom} ({package.payload.messagepayload})")
                        got_packages_from.append(package.header.messagefrom)
                    except queue.Empty:
                        break

                if random.random() <= self.communication_on_active_prob:
                    # send package to a random friend..

                    _alive_ones = [n for n in self.context["alive_nodes"] if n != self.componentinstancenumber]

                    if len(_alive_ones) == 0: # everyone is dead!!!
                        # print(f"  **ROOT** N-{self.componentinstancenumber}: Eveyone is dead!!!")
                        # return None, None       # time to go!
                        to_friend = None
                    else:
                        to_friend = random.choice(_alive_ones)
                        self.send_random_basic_message(to_friend)

                self.alive_for_next_ticks -= 1

                if self.alive_for_next_ticks == 0:
                    if len(got_packages_from) > 0:
                        # got a package, this means immeiate activeness from passive!
                        next_state = DSAHCNodeSimulationStatus.ACTIVE
                        self.alive_for_next_ticks = random.randint(self.min_activeness_after_receive, self.max_activeness_after_receive)
                    else:
                        next_state = DSAHCNodeSimulationStatus.PASSIVE
                else:
                    next_state = DSAHCNodeSimulationStatus.ACTIVE

        assert next_state is not None

        # ST: state
        # NS: next state
        # GPF: got packages from
        # SPF: sent package to friend
        # ANT: alive next ticks
        # P2P: packages to process
        # print(f"   {'ROOT' if self._i_am_root else '==>'} N-{self.componentinstancenumber}: P: {self._parent_node}, CC: ({self._child_counter}) {self._children}, ST: {self.simulation_state}, NS: {next_state}, GPF: {got_packages_from}, SPF: {to_friend}, ANT: {self.alive_for_next_ticks}, P2P: {self.basic_message_queue.qsize()}")

        if self._i_am_root:
            print(f"   {'ROOT' if self._i_am_root else '==>'} N-{self.componentinstancenumber}: P: {self._parent_node}, CC: ({self._child_counter}) {self._children}, ST: {self.simulation_state}, NS: {next_state}, GPF: {got_packages_from}, SPF: {to_friend}, ANT: {self.alive_for_next_ticks}, P2P: {self.basic_message_queue.qsize()}")

        # time.sleep(self.sleep_ms_per_tick / 1000)
        self.__tick_n += 1
        self.simulation_state = next_state

        if self.simulation_state == DSAHCNodeSimulationStatus.PASSIVE:
            self._passive_counter += 1
        elif self.simulation_state == DSAHCNodeSimulationStatus.ACTIVE:
            self._passive_counter = 0

        __cms =  self._cms
        self._cms = 0

        return next_state, to_friend, __cms

class DijkstraScholtenAdHocNode(ComponentModel):
    def __init__(self, componentname, componentid, context):
        self.context = context
        # SUBCOMPONENTS
        self.appllayer = DijkstraScholtenApplicationLayerComponent("ApplicationLayer", componentid, context=self.context)
        self.netlayer = AllSeingEyeNetworkLayer("NetworkLayer", componentid)
        self.linklayer = LinkLayer("LinkLayer", componentid)
        # self.failuredetect = GenericFailureDetector("FailureDetector", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.appllayer.connect_me_to_component(ConnectorTypes.DOWN, self.netlayer)
        # self.failuredetect.connectMeToComponent(PortNames.DOWN, self.netlayer)
        self.netlayer.connect_me_to_component(ConnectorTypes.UP, self.appllayer)
        # self.netlayer.connectMeToComponent(PortNames.UP, self.failuredetect)
        self.netlayer.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.netlayer)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        super().__init__(componentname, componentid, context=self.context)

    def on_init(self, eventobj: Event):
        # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
        pass

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def simulation_tick(self):
        return self.appllayer.simulation_tick()

    @property
    def waiting_packages_on_queue(self):
        if self.appllayer.simulation_state == DSAHCNodeSimulationStatus.OUT_OF_CLOCK or self.appllayer.simulation_state == DSAHCNodeSimulationStatus.OUT_OF_TREE:
            return 0

        return self.appllayer.basic_message_queue.qsize()

    @property
    def parent_node(self):
        return self.appllayer._parent_node


from ahc.Ahc import ComponentModel, Event, GenericMessage, ConnectorTypes, GenericMessageHeader, EventTypes, \
    ComponentRegistry, Lock, Thread, Topology
from ahc.Routing.CGSR.CGSR import CGSRRouter
from ahc.Routing.CGSR.CGSRControlComponent import CGSRControlComponent
from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.CGSR.CGSRDataSender import DataSender

class CGSRNode(ComponentModel):
    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
    def __init__(self, componentname, componentid):
        super(CGSRNode, self).__init__(componentname, componentid)
        self.CGSRControlComponent = CGSRControlComponent(CGSRNode.__name__, componentid)
        self.CGSRRouter = CGSRRouter(CGSRNode.__name__, componentid)
        self.DataSender = DataSender(DataSender.__name__, componentid)

        self.CGSRControlComponent.connect_me_to_component(ConnectorTypes.PEER, self.CGSRRouter)
        self.CGSRRouter.connect_me_to_component(ConnectorTypes.PEER, self.CGSRControlComponent)

        self.CGSRRouter.connect_me_to_component(ConnectorTypes.DOWN, self.DataSender)
        self.DataSender.connect_me_to_component(ConnectorTypes.UP, self.CGSRRouter)

        self.DataSender.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.DataSender)

        # super(CGSRNode, self).__init__(componentname, componentid)
        super().__init__(componentname, componentid)




from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes

class Coordinator(ComponentModel):
    def __init__(self, componentname, componentid):
        super(Coordinator, self).__init__(componentname, componentid)
        self.RoutingTable = {}
        self.Response_Record = {}

    def on_message_from_top(self, eventobj: Event):
        sender = eventobj.eventcontent.header.messagefrom.split("-")[0]
        messageto = eventobj.eventcontent.header.messageto.split("-")[0]
        message_type = eventobj.eventcontent.header.messagetype
        message = eventobj.eventcontent.payload
        print(f"Coordinator receives message from top {messageto} {message_type} {message}")
        if messageto == Coordinator.__name__:
            if sender == "ApplicationComponent" and message_type == "INITIATE":
                if self.componentinstancenumber == 0:
                    message_header = GenericMessageHeader("INITIATEBFSCONSTRUCTION",
                                                          "Coordinator-" + str(self.componentinstancenumber),
                                                          "FredericksonAlgorithmSimpleComponent-" + str(self.componentinstancenumber))
                    message = GenericMessage(message_header, "")
                    kickstarter = Event(self, EventTypes.MFRP, message)
                    self.send_peer(kickstarter)
                    print("Coordinator -> Frederickson")
            elif sender == "ApplicationComponent" and (message_type == "APPQUERY" or message_type == "APPRESPONSE"):
                if message_type == "APPRESPONSE" or len(self.RoutingTable) > 0:
                    dest, info = message
                    path_to_follow = None
                    for i in self.RoutingTable:
                        if dest in i:
                            path_to_follow = i
                            break
                    if message_type == "APPRESPONSE":
                        path_to_follow = self.Response_Record[dest]

                    neighbor_id = path_to_follow[1]
                    if message_type == "APPRESPONSE":
                        print("Neighbor : ", neighbor_id)


                    message_header = GenericMessageHeader(message_type, Coordinator.__name__ + "-" + str(
                        self.componentinstancenumber),
                                                          Coordinator.__name__ + "-" + str(neighbor_id),
                                                          interfaceid=str(self.componentinstancenumber) + "-" + str(
                                                              neighbor_id))
                    mess_ = GenericMessage(message_header, (path_to_follow[1:], (path_to_follow), dest, self.componentinstancenumber, info))

                    event = Event(self, EventTypes.MFRT, mess_)
                    self.send_down(event)
                    print(f"Coordinator {self.componentinstancenumber} sends APPQUERY {neighbor_id} to relay it {dest} - {self.RoutingTable}")


    def on_message_from_peer(self, eventobj: Event):
        sender = eventobj.eventcontent.header.messagefrom.split("-")[0]
        messageto = eventobj.eventcontent.header.messageto.split("-")[0]
        message_type = eventobj.eventcontent.header.messagetype
        message = eventobj.eventcontent.payload
        if messageto == Coordinator.__name__:
            if sender == "FredericksonAlgorithmSimpleComponent" and message_type == "BFSTREECONSTRUCTED":
                self.RoutingTable = message
                print(f"Coordinator {self.componentinstancenumber} has received BFS Tree {self.RoutingTable}")


    def on_init(self, eventobj: Event):
        pass
    def on_message_from_bottom(self, eventobj: Event):
        sender = eventobj.eventcontent.header.messagefrom.split("-")[0]
        messageto = eventobj.eventcontent.header.messageto.split("-")[0]
        message_type = eventobj.eventcontent.header.messagetype
        message = eventobj.eventcontent.payload
        if messageto == Coordinator.__name__:
            if sender == "Coordinator" and (message_type == "APPQUERY" or message_type == "APPRESPONSE"):
                curr_, path_to_follow, dest, source, content = message
                print(f"Coordinator {self.componentinstancenumber} has received APPQUERY {curr_, path_to_follow, dest, source}")

                if dest == self.componentinstancenumber:
                    message_header = GenericMessageHeader(message_type,
                                                          "Coordinator-" + str(self.componentinstancenumber),
                                                          "ApplicationComponent-" + str(self.componentinstancenumber))
                    message_ = GenericMessage(message_header, (source, content))
                    kickstarter = Event(self, EventTypes.MFRB, message_)
                    self.send_up(kickstarter)
                    self.Response_Record[source] = list(reversed(path_to_follow))
                    print(f"Response Record : ", self.Response_Record)
                    # send to app layer
                    pass
                else:
                        neighbor_id = curr_[1]

                        message_header = GenericMessageHeader(message_type, Coordinator.__name__ + "-" + str(
                            self.componentinstancenumber),
                                                              Coordinator.__name__ + "-" + str(neighbor_id),
                                                              interfaceid=str(self.componentinstancenumber) + "-" + str(
                                                                  neighbor_id))
                        mess_ = GenericMessage(message_header, (curr_[1:], path_to_follow, dest, source, content))

                        event = Event(self, EventTypes.MFRT, mess_)
                        self.send_down(event)
                        print(f"*****Routing from {self.componentinstancenumber} to {neighbor_id} - {self.RoutingTable}*****")





#TODO: Component type and hardwired instance numbers are to be corrected
#TODO: Not a good implementation, to be removed
from enum import Enum
import time
from ahc.Ahc import ComponentModel, Event, GenericMessageHeader, GenericMessagePayload, GenericMessage, EventTypes
from threading import Thread

node_number = 2
waiting_ticket = [True]*node_number
ticket_values = [0]*node_number
    
class BakeryMessageTypes(Enum):
  INC = "+"
  DEC = "-"

class ResourceComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}\n")
    pass

  def on_message_from_bottom(self, eventobj: Event):
    messagetype = eventobj.eventcontent.header.messagetype ## INC / DEC
    if messagetype == BakeryMessageTypes.INC:
          print("Increment value")
          self.value += 1
    elif messagetype == BakeryMessageTypes.DEC:
          print("Decrement value")
          self.value -= 1
          
  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.value = 0

  def __repr__(self):
    return f"Value:{self.value}"

class MutualExclusionBakeryComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}\n")
    pass

  def send_message(self):
    global flag,turn
    message_header = GenericMessageHeader(self.messageType, self.componentinstancenumber, None,interfaceid=f"{self.componentinstancenumber}-{self.resourceComponentId}")
    message_payload = GenericMessagePayload(None)
    message = GenericMessage(message_header, message_payload)
          
    ticket_values[self.id] = max(ticket_values)+1
    waiting_ticket[self.id] = False
    for i in range(node_number):
      while waiting_ticket[i] == True:
          continue
      while ticket_values[i] > 0 and (ticket_values[i] < ticket_values[self.id] or(ticket_values[i] == ticket_values[self.id] and i < self.id)):
          continue
    ####critical section####
    self.send_down(Event(self, EventTypes.MFRT, message))        
    time.sleep(0.001)
    ########################
    ticket_values[self.id] = 0    
  
    self.done = True

  def __init__(self, componentname, componentinstancenumber,type = BakeryMessageTypes.INC , resourceComponentId = 2):
    super().__init__(componentname, componentinstancenumber)
    self.resourceComponentId = resourceComponentId
    self.done = False
    self.messageType = type
    self.id = self.componentinstancenumber

  def start(self):
    t = Thread(target=self.send_message)
    t.daemon = True
    t.start()



# given a map of one-hop nodes to two-hop
# neighbors; returns a set of mpr nodes.
def obtain_mpr(valid_bidir_map):
    # initially; all one-hop neighbors are MPR candidates.
    mpr_candidates = set(valid_bidir_map.keys())

    # OLSR allows suboptimal MPR sets; so the simplest
    # MPR selection algorithm would be to return here.
    # but of course, we can do better

    # throughout the algorithm, maintain a set of
    # two-hop nodes that are not yet covered by the
    # already-selected set of MPR nodes.
    # at the start; this is all two-hop nodes,
    # as no mpr nodes are selected yet.
    uncovered_two_hops = all_two_hops_for(mpr_candidates, valid_bidir_map)

    # as the low-hanging fruit; we can start by picking
    # one-hop nodes that have unique access to certain
    # two-hop nodes. that is; if node X is the only one-hop
    # neighbor with access to a two-hop neighbor node Y,
    # then node X simply cannot be excluded from the MPR set.
    # so we start by finding and adding these nodes.
    mpr = one_hops_with_unique_access(valid_bidir_map)
    # remove selected MPR nodes from candidates;
    # also remove their two-hop coverage from uncovered set
    mpr_candidates.difference_update(mpr)
    uncovered_two_hops.difference_update(all_two_hops_for(mpr, valid_bidir_map))

    # this part is greedy, and not guaranteed to give us the
    # most optimal set. however, I think it is a good heuristic
    # to cover most cases:
    # every iteration; we pick the one-hop neighbor that covers
    # most of the uncovered two-hop neighbors.
    # we must iterate until uncovered_two_hops is depleted
    while len(uncovered_two_hops) > 0:
        best_one_hop = max_coverage_neighbor(uncovered_two_hops, valid_bidir_map)
        if best_one_hop is not None:
            mpr.add(best_one_hop)
            # remove selected MPR node from candidates;
            # also remove its two-hop coverage from uncovered set
            mpr_candidates.discard(best_one_hop)
            uncovered_two_hops.difference_update(all_two_hops_for({best_one_hop}, valid_bidir_map))
    # validate the MPR set, just in case
    if not is_mpr_valid(mpr, valid_bidir_map):
        print("MPR selection does not work; reverting to the whole set")
        return set(valid_bidir_map.keys())
    return mpr


# validates that the selected MPR set covers
# all two-hop neighbors as the whole one-hop set
def is_mpr_valid(mpr_set, valid_bidir_map):
    all_one_hops = set(valid_bidir_map.keys())
    if all_one_hops == mpr_set:
        return True
    all_coverage = set()
    mpr_coverage = set()
    for one_hop in all_one_hops:
        all_coverage = all_coverage.union(valid_bidir_map[one_hop])
    for one_hop in mpr_set:
        mpr_coverage = mpr_coverage.union(valid_bidir_map[one_hop])
    return mpr_coverage.difference(all_one_hops) == all_coverage.difference(all_one_hops)


# returns the superset of two-hop nodes
# covered by a set of given one-hop nodes
def all_two_hops_for(one_hops, valid_bidir_map):
    two_hops = set()
    for one_hop in one_hops:
        if one_hop in valid_bidir_map:
            two_hops = two_hops.union(valid_bidir_map[one_hop])
    # exclude nodes that are already one-hop neighbors
    # we want a set of exclusively two-hop neighbors here
    for one_hop in valid_bidir_map.keys():
        two_hops.discard(one_hop)
    return two_hops


# returns the one-hop neighbor that is connected
# to the maximum number of interested two-hop neighbors
def max_coverage_neighbor(interested_two_hops, valid_bidir_map):
    max_coverage = 0
    best_node = None
    for one_hop, two_hops in valid_bidir_map.items():
        coverage = len(interested_two_hops.intersection(two_hops))
        if coverage > max_coverage:
            max_coverage = coverage
            best_node = one_hop
    return best_node


# returns the set of nodes with unique access to certain two-top nodes.
# that is, if two-hop node B is accessible only by one-hop node A; this
# func returns node A.
def one_hops_with_unique_access(valid_bidir_map):
    one_hop_with_uniq = set()
    for two_hop, one_hops in invert_bidir_map(valid_bidir_map).items():
        if len(one_hops) == 0:
            one_hop_with_uniq.add(list(one_hops)[0])
    return one_hop_with_uniq


# input map: one-hop neighbor -> list of two-hop neighbors it connects to
# output map: two-hop neighbor -> list of one-hop neighbors connected to it
def invert_bidir_map(valid_bidir_map):
    inv = {}
    for one_hop, two_hops in valid_bidir_map.items():
        for two_hop in two_hops:
            if two_hop not in inv:
                inv[two_hop] = set()
            inv[two_hop].add(one_hop)
    return inv



from ahc.Ahc import ComponentModel, Event, GenericMessage, ConnectorTypes, GenericMessageHeader, EventTypes, \
    ComponentRegistry, Lock, Thread, Topology
from ahc.Routing.OLSR.OLSRSimplifiedComponent import OLSRSimplifiedRouter
from ahc.Routing.OLSR.OLSRPacketSender import OLSRPacketSender
from ahc.Routing.OLSR.OLSRMessageGenerator import OLSRControlComponent

class OLSRNode(ComponentModel):
    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))
    def __init__(self, componentname, componentid):
        super(OLSRNode, self).__init__(componentname, componentid)
        self.OLSRControlComponent = OLSRControlComponent(OLSRControlComponent.__name__, componentid)
        self.OLSRSimplifiedRouter = OLSRSimplifiedRouter(OLSRSimplifiedRouter.__name__, componentid)
        self.OLSRPacketSender = OLSRPacketSender(OLSRPacketSender.__name__, componentid)


        self.OLSRControlComponent.connect_me_to_component(ConnectorTypes.PEER, self.OLSRSimplifiedRouter)
        self.OLSRSimplifiedRouter.connect_me_to_component(ConnectorTypes.PEER, self.OLSRControlComponent)

        self.OLSRSimplifiedRouter.connect_me_to_component(ConnectorTypes.DOWN, self.OLSRPacketSender)
        self.OLSRPacketSender.connect_me_to_component(ConnectorTypes.UP, self.OLSRSimplifiedRouter)

        self.OLSRPacketSender.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.OLSRPacketSender)

        super().__init__(componentname, componentid)




import threading
import logging
import time
from copy import deepcopy

from networkx import Graph
from typing import Dict, List, Any, Tuple

from ahc.Ahc import *
from ahc.Routing.STAR.MinHeap import MinHeap, MinHeapNode
from ahc.Routing.STAR.helper import STARStats, STARStatEvent

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s [%(levelname)s] - %(message)s')
logging.getLogger("matplotlib").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


class STARMessageTypes(Enum):
    NEIGHBOR_UP = "NeighborUp"
    LSU = "LinkStateUpdate"
    APP = "AppMessage"


class LSUMessage:
    def __init__(self, src: int, dest: int, link_cost: float, timestamp: int):
        self.src: int = src
        self.dest: int = dest
        self.link_cost: float = link_cost
        self.time: int = timestamp

    def __str__(self) -> str:
        return f"({self.src}, {self.dest}, {self.link_cost}, {self.time})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class STARNodeComponent(ComponentModel):
    def __init__(self, componentname, componentinstancenumber):
        """ Node initializes itself """
        super().__init__(componentname, componentinstancenumber)
        self.topology_graphs: Dict = {self.componentinstancenumber: nx.Graph()}
        self.source_trees: Dict = {self.componentinstancenumber: nx.DiGraph()}
        self.routing_table = {}
        self.neighbors: List[int] = []
        self.messages: List[Tuple[STARMessageTypes, Any]] = []
        self.message_handlers: Dict = {
            STARMessageTypes.NEIGHBOR_UP: self.on_neighbor_up,
            STARMessageTypes.LSU: self.on_lsu,
            STARMessageTypes.APP: self.on_app_message
        }
        self.stats = STARStats()

    def on_init(self, eventobj: Event):
        """ Node tells its neighbors it is up """
        logger.info(f"STAR Node init: {self.componentinstancenumber}. ")
        logger.debug(f"T{threading.get_native_id()} # {self.componentinstancenumber} is announcing")

        self.announce_node_up()

    def on_message_from_bottom(self, eventobj: Event):
        event_type = eventobj.eventcontent.header.messagetype
        message_from = eventobj.eventcontent.header.messagefrom
        payload = eventobj.eventcontent.payload.messagepayload

        logger.debug(f"#{self.componentinstancenumber} got message from {message_from} = {payload}")

        self.message_handlers[event_type](eventobj)

    def on_app_message(self, eventobj: Event):
        message = eventobj.eventcontent
        message_to = message.header.messageto

        if message_to == self.componentinstancenumber:
            # incoming message is for me
            self.stats.push(STARStatEvent.APP_MSG_RECV, message.payload.messagepayload)

            self.send_up(Event(self, EventTypes.MFRB, message))
        else:
            try:
                next_hop = self.routing_table[message_to][0]

                message.header.nexthop = next_hop
                logger.debug(f"T{threading.get_native_id()} #{self.componentinstancenumber} Next hop: {next_hop}")
                message.payload.messagepayload['hop_count'] += 1
                self.send_down(Event(self, EventTypes.MFRT, message))
            except KeyError:
                # destination is not existed in the routing table
                # drop the message
                logger.error(f'Destination {message_to} not found in node {self.componentinstancenumber}')

    def on_message_from_top(self, eventobj: Event):
        message = eventobj.eventcontent
        message_to = message.header.messageto

        try:
            next_hop = self.routing_table[message_to][0]
            message.header.nexthop = next_hop
            self.send_down(Event(self, EventTypes.MFRT, message))
        except KeyError:
            # destination is not existed in the routing table
            # drop the message
            logger.error(f'Destination {message_to} not found in node {self.componentinstancenumber}')

        self.stats.push(STARStatEvent.APP_MSG_SENT)

    def announce_node_up(self):
        payload = {"node_id": self.componentinstancenumber}
        self.add_message(STARMessageTypes.NEIGHBOR_UP, payload)
        self.broadcast_messages()

    def add_message(self, message_type: STARMessageTypes, payload: Any):
        msg = (message_type, payload)
        self.messages.append(msg)

    def broadcast_messages(self):
        if len(self.messages) > 0:
            message_type = self.messages[0][0]
            hdr = GenericMessageHeader(message_type, self.componentinstancenumber,
                                       MessageDestinationIdentifiers.NETWORKLAYERBROADCAST)
            payload = GenericMessagePayload(self.messages)

            if message_type == STARMessageTypes.LSU:
                self.stats.push(STARStatEvent.LSU_MSG_SENT, len(self.messages))
                self.stats.push(STARStatEvent.UPDATE_MSG_SENT, 1)

            self.send_down(Event(self, EventTypes.MFRT, GenericMessage(hdr, payload)))

        self.messages = []

    def on_neighbor_up(self, eventobj: Event):
        """ Node gets a NEIGHBOR_UP message and update its source tree and topology graph, and announce link update """

        message_from = eventobj.eventcontent.header.messagefrom
        payload = eventobj.eventcontent.payload.messagepayload

        if len(payload) == 0:
            return
        elif len(payload) == 1:
            node_id = message_from
            link_cost = Topology().G[node_id][self.componentinstancenumber]['weight']
        else:
            raise NotImplementedError()

        self.neighbors.append(node_id)
        self.topology_graphs[node_id] = nx.Graph()
        self.source_trees[node_id] = nx.DiGraph()
        send_st = True  # Report source tree changes

        # TODO LORA

        # Update
        lsu = LSUMessage(self.componentinstancenumber, node_id, link_cost, self.__now())
        self.update(self.componentinstancenumber, [lsu])

        if send_st:
            my_st: Graph = self.source_trees[self.componentinstancenumber]
            my_tg: Graph = self.topology_graphs[self.componentinstancenumber]
            self.messages = []

            for (u, v) in my_st.edges:
                l, t = my_tg[u][v]['weight'], my_tg[u][v]['time']
                lsu = LSUMessage(u, v, l, t)
                logger.debug(f"T{threading.get_native_id()} # {self.componentinstancenumber} is sending lsu {lsu}")
                self.add_message(STARMessageTypes.LSU, lsu)

            self.broadcast_messages()

    def on_lsu(self, eventobj: Event):
        self.stats.push(STARStatEvent.LSU_MSG_RECV, 1)
        message_from = eventobj.eventcontent.header.messagefrom
        payload = eventobj.eventcontent.payload.messagepayload

        # logger.debug(f"T{threading.get_native_id()} # {self.componentinstancenumber} is updating from lsu {payload}")

        self.update(message_from, payload)

    def update(self, k: int, payload: List[LSUMessage]):
        """ Process update message 'lsu' sent by router 'node_id' """
        self.update_topology_graph(k, payload)

        if k != self.componentinstancenumber:
            self.build_shortest_path_tree(k)

        self.build_shortest_path_tree(self.componentinstancenumber)
        self.update_routing_table()

        if k != self.componentinstancenumber:
            self.broadcast_messages()

    def update_topology_graph(self, k: int, payload: List[LSUMessage]):
        for lsu in payload:
            if type(lsu) == tuple:
                lsu = lsu[1]
            if lsu.link_cost != float("inf"):
                self.process_add_update(k, lsu)
            else:
                self.process_void_update(k, lsu)

    def build_shortest_path_tree(self, k: int):
        """ Construct source tree of node k """
        self.init_single_source(k)
        k_tg: Graph = self.topology_graphs[k]
        my_tg: Graph = self.topology_graphs[self.componentinstancenumber]
        min_heap = MinHeap()
        new_st = nx.DiGraph()

        for node in k_tg.nodes:
            min_heap.insert(MinHeapNode(node, k_tg.nodes[node]))

        popped_node: MinHeapNode = min_heap.extract_min()
        d, u, u_data = popped_node.data['d'], popped_node.key, popped_node.data

        while u_data is not None and u_data['d'] < float('inf'):
            pred = u_data['pred']

            if pred is not None and not new_st.has_edge(pred[0], pred[1]):
                new_st.add_edge(pred[0], pred[1], weight=pred[2]['weight'], time=pred[2]['time'], dlt=pred[2]['dlt'])

            adj_list = {n: nbrdict for n, nbrdict in k_tg.adjacency()}

            for v in adj_list[u]:
                cond_1 = k_tg.has_edge(u, v) or k_tg[u][v]['weight'] != float('inf')
                cond_2 = my_tg.has_edge(u, v) and not my_tg[u][v]['dlt']

                if cond_1 and cond_2:
                    pass
                else:
                    continue

                if k == self.componentinstancenumber:
                    if u == self.componentinstancenumber:
                        suc = self.componentinstancenumber
                    elif my_tg.nodes[u]['suc'] == self.componentinstancenumber:
                        suc = u
                    else:
                        suc = my_tg.nodes[u]['suc']
                else:
                    if u == k:
                        if v == self.componentinstancenumber:
                            suc = self.componentinstancenumber
                        else:
                            suc = k
                    else:
                        try:
                            suc = my_tg.nodes[u]['suc']
                        except KeyError:
                            suc = None

                self.relax_edge(k, u, v, min_heap, suc)

            if not min_heap.is_empty():
                popped_node: MinHeapNode = min_heap.extract_min()
                d, u, u_data = popped_node.data['d'], popped_node.key, popped_node.data
            else:
                d, u, u_data = None, None, None

        self.update_neighbor_tree(k, new_st)

        if k == self.componentinstancenumber:
            # TODO LORA
            self.report_changes(self.source_trees[self.componentinstancenumber], new_st)
            my_tg_edges = deepcopy(my_tg.edges)

            for (u, v) in my_tg_edges:
                if my_tg[u][v]['dlt']:
                    my_tg.remove_edge(u, v)

        self.source_trees[k] = new_st

    def update_routing_table(self):
        my_st = self.source_trees[self.componentinstancenumber]

        for dest in my_st.nodes:
            if dest != self.componentinstancenumber:
                path = nx.dijkstra_path(my_st, self.componentinstancenumber, dest)
                length = nx.dijkstra_path_length(my_st, self.componentinstancenumber, dest)
                self.routing_table[dest] = (path[1], length)

    def process_add_update(self, k: int, lsu: LSUMessage):
        """ Update topology graphs of this node and the neighbor from 'lsu' """
        my_tg: Graph = self.topology_graphs[self.componentinstancenumber]
        edge_existed = my_tg.has_edge(lsu.src, lsu.dest)

        if not edge_existed or lsu.time > my_tg[lsu.src][lsu.dest]['time']:
            if not edge_existed:
                my_tg.add_edge(lsu.src, lsu.dest, weight=lsu.link_cost, time=lsu.time, dlt=False)
            else:
                my_tg[lsu.src][lsu.dest]['weight'] = lsu.link_cost
                my_tg[lsu.src][lsu.dest]['time'] = lsu.time

        if k != self.componentinstancenumber:
            k_tg: Graph = self.topology_graphs[k]
            k_tg_edges = deepcopy(k_tg.edges)
            for (r, s) in k_tg_edges:
                if r != lsu.src and s == lsu.dest:
                    k_tg.remove_edge(r, s)

            if not k_tg.has_edge(lsu.src, lsu.dest):
                k_tg.add_edge(lsu.src, lsu.dest, weight=lsu.link_cost, time=lsu.time, dlt=False)
            else:
                k_tg[lsu.src][lsu.dest]['weight'] = lsu.link_cost
                k_tg[lsu.src][lsu.dest]['time'] = lsu.time

        my_tg[lsu.src][lsu.dest]['dlt'] = False

    def process_void_update(self, k: int, lsu: LSUMessage):
        # TODO on link failure
        pass

    def init_single_source(self, k: int):
        k_tg: Graph = self.topology_graphs[k]

        for v in k_tg.nodes:
            k_tg.nodes[v]['d'] = float('inf')
            k_tg.nodes[v]['pred'] = None
            try:
                k_tg.nodes[v]['sucp'] = k_tg.nodes[v]['suc']
            except KeyError:
                k_tg.nodes[v]['sucp'] = None
            k_tg.nodes[v]['suc'] = None
            k_tg.nodes[v]['nbr'] = None

        k_tg.nodes[k]['d'] = 0

    def relax_edge(self, k: int, u: int, v: int, min_heap: MinHeap, suc: int):
        k_tg: Graph = self.topology_graphs[k]
        my_st: Graph = self.source_trees[self.componentinstancenumber]

        COND_1 = k_tg.nodes[v]['d'] > k_tg.nodes[u]['d'] + k_tg[u][v]['weight']

        COND_2 = k == self.componentinstancenumber and \
                 k_tg.nodes[v]['d'] == k_tg.nodes[u]['d'] + k_tg[u][v]['weight'] and \
                 my_st.has_edge(u, v)

        if COND_1 or COND_2:
            k_tg.nodes[v]['d'] = k_tg.nodes[u]['d'] + k_tg[u][v]['weight']
            k_tg.nodes[v]['pred'] = (u, v, k_tg[u][v])
            k_tg.nodes[v]['suc'] = suc

            # TODO LORA

            min_heap.insert(MinHeapNode(v, k_tg.nodes[v]))

    def update_neighbor_tree(self, k: int, new_st: Graph):
        """ Delete links from k's topology graph and report failed links"""
        k_tg: Graph = self.topology_graphs[k]
        k_st: Graph = self.source_trees[k]
        k_st_edges = deepcopy(k_st.edges)
        my_tg: Graph = self.topology_graphs[self.componentinstancenumber]

        for (u, v) in k_st_edges:
            if not new_st.has_edge(u, v):
                # k has removed (u, v) from its source tree
                # TODO LORA
                if k == self.componentinstancenumber and \
                        (u == self.componentinstancenumber or my_tg.nodes[v]['pred'] is None):
                    # i has no path to destination v or i is the head node
                    if my_tg.nodes[v]['pred'] is None:
                        for (r, s) in my_tg.edges:
                            if s == v:
                                if not my_tg.has_edge(r, s) or my_tg[r][s]['weight'] == float('inf'):
                                    l = my_tg[r][s]['weight']
                                    t = my_tg[r][s]['time']
                                    lsu = LSUMessage(r, s, l, t)
                                    self.add_message(STARMessageTypes.LSU, lsu)
                    elif not my_tg.has_edge(u, v) or my_tg[u][v]['weight'] == float('inf'):
                        l = my_tg[u][v]['weight']
                        t = my_tg[u][v]['time']
                        lsu = LSUMessage(u, v, l, t)
                        self.add_message(STARMessageTypes.LSU, lsu)
                if not (k == self.componentinstancenumber and u == self.componentinstancenumber):
                    if k_tg.has_edge(u, v):
                        k_tg.remove_edge(u, v)

                    if my_tg.has_edge(u, v) and my_tg[u][v]['weight'] != float('inf'):
                        for x in self.neighbors:
                            x_tg: Graph = self.topology_graphs[x]
                            if x_tg.has_edge(u, v):
                                return
                        my_tg[u][v]['dlt'] = True

    def report_changes(self, old_st: Graph, new_st: Graph):
        """ Generate new LSUs for new links in the router's source tree """
        my_tg: Graph = self.topology_graphs[self.componentinstancenumber]

        for (u, v) in new_st.edges:
            if not old_st.has_edge(u, v) or \
                    new_st[u][v]['time'] != old_st[u][v]['time']:
                try:
                    l = my_tg[u][v]['weight']
                    t = my_tg[u][v]['time']

                    lsu = LSUMessage(u, v, l, t)
                    self.add_message(STARMessageTypes.LSU, lsu)
                except KeyError:
                    pass

    def __now(self):
        return time.time_ns() // 1000

    def plot(self, gg: Graph):
        import matplotlib.pyplot as plt
        nodepos = nx.drawing.spring_layout(gg)
        nodecolors = ['#6aff20'] * len(gg.nodes)
        nx.draw(gg, nodepos, node_color=nodecolors, with_labels=True, font_weight='bold')

        labels = nx.get_edge_attributes(gg, 'weight')
        nx.draw_networkx_edge_labels(gg, nodepos, edge_labels=labels)

        plt.show()

    def show_routing_table(self):
        print(f'{self.componentinstancenumber} => {self.routing_table}')

    def show_topology_graph(self, k):
        self.plot(self.topology_graphs[k])

    def show_source_tree(self, k):
        self.plot(self.source_trees[k])



import logging
import statistics
import sys
import threading
from ahc.Consensus.Raft.log import LogManager

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

logger = logging.getLogger(__name__)

class State:
    

    def __init__(self, old_state=None, server=None):

        if old_state:
            self.server = old_state.server
            self.votedFor = old_state.votedFor
            self.currentTerm = old_state.currentTerm
            self.leaderId = old_state.leaderId
            self.log = old_state.log
        else:
            self.server = server
            self.votedFor = None
            self.currentTerm = 0
            self.leaderId = None
            self.log = LogManager()

    def data_received_peer(self, peer, msg):

        logger.debug('For component %s Received %s from %s', msg['type'], self.server.componentinstancenumber ,peer.componentinstancenumber)

        if self.currentTerm < msg['term']:
            self.currentTerm = msg['term']
            if not type(self) is Follower:
                logger.info('For component %s Remote term is higher, converting to Follower', self.server.componentinstancenumber)
                self.server.change_state(Follower)
                self.server.state.data_received_peer(peer, msg)
                return
        method = getattr(self, 'on_peer_' + msg['type'], None)
        if method:
            method(peer, msg)
        else:
            logger.info('For component %s Unrecognized message from %s: %s', self.server.componentinstancenumber,peer.componentinstancenumber, msg)

    def data_received_client(self, client, msg):

        method = getattr(self, 'on_client_' + msg['type'], None)
        if method:
            method(client, msg)

    def on_client_append(self, client, msg):
        
        msg = {'type': 'redirect',
               'leader': self.leaderId}
        client.send(msg)
        logger.debug('Redirect client %s to leader',
                     self.leaderId)

    def on_client_get(self, client, msg):
        
        client.send(self.log)


class Follower(State):
    

    def __init__(self, old_state=None, server=None):
        
        super().__init__(old_state, server)
        self.votedFor = None
        self.restart_election_timer()

    def teardown(self):
        
        self.election_timer.cancel()

    def restart_election_timer(self):
        
        if hasattr(self, 'election_timer'):
            self.election_timer.cancel()

        timeout = ord(self.server.componentinstancenumber) - ord('A') + 1
        #loop = get_or_create_eventloop()
        #self.election_timer = loop. \
        #    call_later(timeout, self.server.change_state, Candidate)
        self.election_timer = threading.Timer(timeout, self.server.change_state, (Candidate,))
        self.election_timer.start()
        logger.debug(' For component %s Election timer restarted: %s s', self.server.componentinstancenumber,timeout)

    def on_peer_request_vote(self, peer, msg):
        
        term_is_current = msg['term'] >= self.currentTerm
        can_vote = self.votedFor is None
        index_is_current = (msg['lastLogTerm'] > self.log.term() or
                            (msg['lastLogTerm'] == self.log.term() and
                             msg['lastLogIndex'] >= self.log.index))
        granted = term_is_current and can_vote and index_is_current

        if granted:
            self.votedFor = msg['candidateId']
            self.restart_election_timer()

        logger.debug('For component %s Voting for %s. Term:%s Vote:%s Index:%s', self.server.componentinstancenumber,
                     peer.componentinstancenumber, term_is_current, can_vote, index_is_current)

        response = {'type': 'response_vote', 'voteGranted': granted,
                    'term': self.currentTerm}
        self.server.send_to_component(peer, response)

    def on_peer_append_entries(self, peer, msg):

        self.restart_election_timer()

        term_is_current = msg['term'] >= self.currentTerm
        prev_log_term_match = msg['prevLogTerm'] is None or \
                              self.log.term(msg['prevLogIndex']) == msg['prevLogTerm']
        success = term_is_current and prev_log_term_match

        if success:
            self.log.append_entries(msg['entries'], msg['prevLogIndex'])
            self.log.commit(msg['leaderCommit'])
            self.leaderId = msg['leaderId']
            logger.debug('For component %s Log index is now %s', self.server.componentinstancenumber, self.log.index)
        else:
            logger.warning('For component %s Could not append entries. cause: %s', self.server.componentinstancenumber,
                           'wrong\
                term' if not term_is_current else 'prev log term mismatch')

        resp = {'type': 'response_append', 'success': success,
                'term': self.currentTerm,
                'matchIndex': self.log.index}
        self.server.send_to_component(peer, resp)


class Candidate(Follower):
    

    def __init__(self, old_state=None, server=None):
        
        super().__init__(old_state, server)
        self.currentTerm += 1
        self.votes_count = 0
        logger.info('New Election. Term: %s', self.currentTerm)
        #TODO: put it in seperate thread
        def vote_self():
            self.votedFor = self.server.componentinstancenumber
            self.on_peer_response_vote(
                self.votedFor, {'voteGranted': True})

        vote_self()
        self.send_vote_requests()



    def send_vote_requests(self):
        
        logger.info(' For component %s Broadcasting request_vote', self.server.componentinstancenumber)
        msg = {'type': 'request_vote', 'term': self.currentTerm,
               'candidateId': self.votedFor,
               'lastLogIndex': self.log.index,
               'lastLogTerm': self.log.term()}
        self.server.broadcast_peers(msg)

    def on_peer_append_entries(self, peer, msg):
        
        logger.debug('For component %s Converting to Follower', self.server.componentinstancenumber)
        self.server.change_state(Follower)
        self.server.state.on_peer_append_entries(peer, msg)

    def on_peer_response_vote(self, peer, msg):
        
        self.votes_count += msg['voteGranted']
        logger.info(' For component %s Vote count: %s', self.server.componentinstancenumber, self.votes_count)
        if self.votes_count > len(self.server.registry.get_non_channel_components()) / 2:
            self.server.change_state(Leader)


class Leader(State):
    

    def __init__(self, old_state=None, server=None):

        super().__init__(old_state, server)
        logger.info('For component %s Leader of term: %s', self.server.componentinstancenumber,self.currentTerm)
        self.leaderId = self.server.componentinstancenumber
        cluster = self.server.registry.get_non_channel_components()
        cluster_names = [c.componentinstancenumber for c in cluster]
        self.matchIndex = {p: 0 for p in cluster_names}
        self.nextIndex = {p: self.log.commitIndex + 1 for p in self.matchIndex}
        self.waiting_clients = {}
        self.send_append_entries()

        self.log.append_entries([{'term': self.currentTerm,
                                  'data': {
                                      'key': 'leaderId',
                                      'value': self.server.componentinstancenumber,
                                      }}],
                                self.log.index)
        self.log.commit(self.log.index)

    def teardown(self):
        
        self.append_timer.cancel()

        for clients in self.waiting_clients.values():
            for client in clients:
                client.send({'type': 'result', 'success': False})
                logger.error('Sent unsuccessful response to client')

    def send_append_entries(self):

        cluster = self.server.registry.get_non_channel_components()
        for peer in cluster:
            if peer.componentinstancenumber == self.server.componentinstancenumber:
                continue
            msg = {'type': 'append_entries',
                   'term': self.currentTerm,
                   'leaderCommit': self.log.commitIndex,
                   'leaderId': self.server.componentinstancenumber,
                   'prevLogIndex': self.nextIndex[peer.componentinstancenumber] - 1,
                   'entries': self.log.log
                   }
            msg.update({'prevLogTerm': self.log.term(msg['prevLogIndex'])})

            logger.debug(' For component %s Sending %s entries to %s. Start index %s',self.server.componentinstancenumber,
                         len(msg['entries']), peer.componentinstancenumber, self.nextIndex[peer.componentinstancenumber])
            self.server.send_to_component(peer, msg)

        timeout = 1
        #loop = get_or_create_eventloop()
        #self.append_timer = loop.call_later(timeout, self.send_append_entries)
        self.append_timer = threading.Timer(timeout, self.send_append_entries)
        self.append_timer.start()

    def on_peer_response_append(self, peer, msg):

        if msg['success']:
            self.matchIndex[peer] = msg['matchIndex']
            self.nextIndex[peer] = msg['matchIndex'] + 1

            self.matchIndex[self.server.componentinstancenumber] = self.log.index
            self.nextIndex[self.server.componentinstancenumber] = self.log.index + 1
            index = statistics.median_low(self.matchIndex.values())
            self.log.commit(index)
            self.send_client_append_response()
        else:
            self.nextIndex[peer.componentinstancenumber] = max(0, self.nextIndex[peer.componentinstancenumber] - 1)

    def on_client_append(self, client, msg):
        
        entry = {'term': self.currentTerm, 'data': msg['data']}

        self.log.append_entries([entry], self.log.index)
        if self.log.index in self.waiting_clients:
            self.waiting_clients[self.log.index].append(client)
        else:
            self.waiting_clients[self.log.index] = [client]
        self.on_peer_response_append(
            self.server.componentinstancenumber, {'success': True,
                                       'matchIndex': self.log.commitIndex})

    def send_client_append_response(self):
        
        to_delete = []
        for client_index, clients in self.waiting_clients.items():
            if client_index <= self.log.commitIndex:
                for client in clients:
                    client.send({'type': 'result', 'success': True})  # TODO
                    logger.debug('Sent successful response to client')
                to_delete.append(client_index)
        for index in to_delete:
            del self.waiting_clients[index]
